import os
import torch
from torch import Tensor
from collections import deque
from tqdm import tqdm
from .patches import Patches
from .utils import *
from .bound_ops import *
import warnings

from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
    from .bound_general import BoundedModule


def batched_backward(self: 'BoundedModule', node, C, unstable_idx, batch_size,
                     bound_lower=True, bound_upper=True, return_A=None):
    if return_A is None: return_A = self.return_A
    crown_batch_size = self.bound_opts['crown_batch_size']
    output_shape = node.output_shape[1:]
    dim = int(prod(output_shape))
    if unstable_idx is None:
        unstable_idx = torch.arange(dim, device=self.device)
        dense = True
    else:
        dense = False
    unstable_size = get_unstable_size(unstable_idx)
    if self.verbose:
        print(f'Batched CROWN: node {node}, unstable size {unstable_size}')
    num_batches = (unstable_size + crown_batch_size - 1) // crown_batch_size
    ret = []
    ret_A = {} # if return_A, we will store A here
    pbar = tqdm(range(num_batches)) if self.verbose else range(num_batches)
    for i in pbar:
        if isinstance(unstable_idx, tuple):
            unstable_idx_batch = tuple(
                u[i*crown_batch_size:(i+1)*crown_batch_size]
                for u in unstable_idx
            )
            unstable_size_batch = len(unstable_idx_batch[0])
        else:
            unstable_idx_batch = unstable_idx[i*crown_batch_size:(i+1)*crown_batch_size]
            unstable_size_batch = len(unstable_idx_batch)
        if node.patches_start and node.mode == "patches":
            assert C in ['Patches', None]
            C_batch = Patches(shape=[
                unstable_size_batch, batch_size, *node.output_shape[1:-2], 1, 1],
                identity=1, unstable_idx=unstable_idx_batch,
                output_shape=[batch_size, *node.output_shape[1:]])
        elif isinstance(node, (BoundLinear, BoundMatMul)):
            assert C in ['OneHot', None]
            C_batch = OneHotC(
                [batch_size, unstable_size_batch, *node.output_shape[1:]],
                self.device, unstable_idx_batch, None)
        else:
            assert C in ['eye', None]
            C_batch = torch.zeros([1, unstable_size_batch, dim], device=self.device)
            C_batch[0, torch.arange(unstable_size_batch), unstable_idx_batch] = 1.0
            C_batch = C_batch.expand(batch_size, -1, -1).view(
                batch_size, unstable_size_batch, *output_shape)
        # overwrite return_A options to run backward general
        ori_return_A_option = self.return_A
        self.return_A = return_A

        batch_ret = self.backward_general(
            bound_node=node, C=C_batch,
            bound_lower=bound_lower, bound_upper=bound_upper,
            average_A=False, need_A_only=False, unstable_idx=unstable_idx_batch,
            verbose=False)
        ret.append(batch_ret[:2])

        if len(batch_ret) > 2:
            # A found, we merge A
            batch_A = batch_ret[2]
            ret_A = merge_A(batch_A, ret_A)

        # restore return_A options
        self.return_A = ori_return_A_option
    if bound_lower:
        lb = torch.cat([item[0].view(batch_size, -1) for item in ret], dim=1)
        if dense:
            # In this case, restore_sparse_bounds will not be called.
            # And thus we restore the shape here.
            lb = lb.reshape(batch_size, *output_shape)
    else:
        lb = None
    if bound_upper:
        ub = torch.cat([item[1].view(batch_size, -1) for item in ret], dim=1)
        if dense:
            # In this case, restore_sparse_bounds will not be called.
            # And thus we restore the shape here.
            ub = ub.reshape(batch_size, *output_shape)
    else:
        ub = None

    if return_A:
        return lb, ub, ret_A
    else:
        return lb, ub


def backward_general(
    self: 'BoundedModule',
    bound_node,
    C,
    start_backpropagation_at_node = None,
    bound_lower=True,
    bound_upper=True,
    average_A=False,
    need_A_only=False,
    unstable_idx=None,
    update_mask=None,
    verbose=True,
    initial_As: Optional[dict] = None,
    initial_lb: Optional[torch.tensor] = None,
    initial_ub: Optional[torch.tensor] = None,
):
    use_beta_crown = self.bound_opts['optimize_bound_args']['enable_beta_crown']

    roots = self.roots()

    if start_backpropagation_at_node is None:
        # When output constraints are used, backward_general_with_output_constraint()
        # adds additional layers at the end, performs the backpropagation through these,
        # and then calls backward_general() on the output layer.
        # In this case, the layer we start from (start_backpropagation_at_node) differs
        # from the layer that should be bounded (bound_node)

        # When output constraints are not used, the bounded node is the one where
        # backpropagation starts.
        start_backpropagation_at_node = bound_node

    if verbose:
        logger.debug(f'Bound backward from {start_backpropagation_at_node.__class__.__name__}({start_backpropagation_at_node.name}) '
                     f'to bound {bound_node.__class__.__name__}({bound_node.name})')
        if isinstance(C, str):
            logger.debug(f'  C: {C}')
        elif C is not None:
            logger.debug(f'  C: shape {C.shape}, type {type(C)}')
    _print_time = bool(os.environ.get('AUTOLIRPA_PRINT_TIME', 0))

    if isinstance(C, str):
        # If C is a str, use batched CROWN. If batched CROWN is not intended to
        # be enabled, C must be a explicitly provided non-str object for this function.
        if need_A_only or average_A:
            raise ValueError(
                'Batched CROWN is not compatible with '
                f'need_A_only={need_A_only}, average_A={average_A}')
        ret = self.batched_backward(
            bound_node, C, unstable_idx,
            batch_size=roots[0].value.shape[0],
            bound_lower=bound_lower, bound_upper=bound_upper,
        )
        bound_node.lower, bound_node.upper = ret[:2]
        return ret

    for n in self.nodes():
        n.lA = n.uA = None

    degree_out = get_degrees(start_backpropagation_at_node)
    C, batch_size, output_dim, output_shape = self._preprocess_C(C, bound_node)

    if initial_As is None:
        start_backpropagation_at_node.lA = C if bound_lower else None
        start_backpropagation_at_node.uA = C if bound_upper else None
    else:
        for layer_name, (lA, uA) in initial_As.items():
            self[layer_name].lA = lA
            self[layer_name].uA = uA
        assert start_backpropagation_at_node.lA is not None or start_backpropagation_at_node.uA is not None
    if initial_lb is None:
        lb = torch.tensor(0., device=self.device)
    else:
        lb = initial_lb
    if initial_ub is None:
        ub = torch.tensor(0., device=self.device)
    else:
        ub = initial_ub

    # Save intermediate layer A matrices when required.
    A_record = {}

    queue = deque([start_backpropagation_at_node])
    while len(queue) > 0:
        l = queue.popleft()  # backward from l
        self.backward_from[l.name].append(bound_node)

        if l.name in self.root_names: continue

        # if all the succeeds are done, then we can turn to this node in the
        # next iteration.
        for l_pre in l.inputs:
            degree_out[l_pre.name] -= 1
            if degree_out[l_pre.name] == 0:
                queue.append(l_pre)

        # Initially, l.lA or l.uA will be set to C for this node.
        if l.lA is not None or l.uA is not None:
            if verbose:
                logger.debug(f'  Bound backward to {l} (out shape {l.output_shape})')
                if l.lA is not None:
                    logger.debug('    lA type %s shape %s',
                                 type(l.lA), list(l.lA.shape))
                if l.uA is not None:
                    logger.debug('    uA type %s shape %s',
                                 type(l.uA), list(l.uA.shape))

            if _print_time:
                start_time = time.time()

            if not l.perturbed:
                if not hasattr(l, 'forward_value'):
                    self.get_forward_value(l)
                lb, ub = add_constant_node(lb, ub, l)
                continue

            if l.zero_uA_mtx and l.zero_lA_mtx:
                # A matrices are all zero, no need to propagate.
                continue

            lA, uA = l.lA, l.uA
            if (l.name != start_backpropagation_at_node.name and use_beta_crown
                    and getattr(l, 'sparse_betas', None)):
                lA, uA, lbias, ubias = self.beta_crown_backward_bound(
                    l, lA, uA, start_node=start_backpropagation_at_node)
                lb = lb + lbias
                ub = ub + ubias

            if isinstance(l, BoundOptimizableActivation):
                # For other optimizable activation functions (TODO: unify with ReLU).
                if bound_node.name != self.final_node_name:
                    start_shape = bound_node.output_shape[1:]
                else:
                    start_shape = C.shape[0]
                l.preserve_mask = update_mask
            else:
                start_shape = None
            A, lower_b, upper_b = l.bound_backward(
                lA, uA, *l.inputs,
                start_node=bound_node, unstable_idx=unstable_idx,
                start_shape=start_shape)

            # After propagation through this node, we delete its lA, uA variables.
            if bound_node.name != self.final_name:
                del l.lA, l.uA
            if _print_time:
                torch.cuda.synchronize()
                time_elapsed = time.time() - start_time
                if time_elapsed > 5e-3:
                    print(l, time_elapsed)
            if lb.ndim > 0 and type(lower_b) == Tensor and self.conv_mode == 'patches':
                lb, ub, lower_b, upper_b = check_patch_biases(lb, ub, lower_b, upper_b)
            lb = lb + lower_b
            ub = ub + upper_b
            if self.return_A and self.needed_A_dict and bound_node.name in self.needed_A_dict:
                # FIXME remove [0][0] and [0][1]?
                if len(self.needed_A_dict[bound_node.name]) == 0 or l.name in self.needed_A_dict[bound_node.name]:
                    # A could be either patches (in this case we cannot transpose so directly return)
                    # or matrix (in this case we transpose)
                    A_record.update({
                        l.name: {
                            "lA": (
                                A[0][0] if isinstance(A[0][0], Patches)
                                else A[0][0].transpose(0, 1).detach()
                            ) if A[0][0] is not None else None,
                            "uA": (
                                A[0][1] if isinstance(A[0][1], Patches)
                                else A[0][1].transpose(0, 1).detach()
                            ) if A[0][1] is not None else None,
                            # When not used, lb or ub is tensor(0).
                            "lbias": lb.transpose(0, 1).detach() if lb.ndim > 1 else None,
                            "ubias": ub.transpose(0, 1).detach() if ub.ndim > 1 else None,
                            "unstable_idx": unstable_idx
                    }})
                # FIXME: solve conflict with the following case
                self.A_dict.update({bound_node.name: A_record})
                if need_A_only and set(self.needed_A_dict[bound_node.name]) == set(A_record.keys()):
                    # We have collected all A matrices we need. We can return now!
                    self.A_dict.update({bound_node.name: A_record})
                    # Do not concretize to save time. We just need the A matrices.
                    # return A matrix as a dict: {node_start.name: [A_lower, A_upper]}
                    return None, None, self.A_dict

            for i, l_pre in enumerate(l.inputs):
                add_bound(l, l_pre, lA=A[i][0], uA=A[i][1])

    if lb.ndim >= 2:
        lb = lb.transpose(0, 1)
    if ub.ndim >= 2:
        ub = ub.transpose(0, 1)

    if self.return_A and self.needed_A_dict and bound_node.name in self.needed_A_dict:
        save_A_record(
            bound_node, A_record, self.A_dict, roots,
            self.needed_A_dict[bound_node.name],
            lb=lb, ub=ub, unstable_idx=unstable_idx)

    lb, ub = concretize(self, batch_size, output_dim, lb, ub,
                        bound_lower, bound_upper,
                        average_A=average_A, node_start=bound_node)

    lb = lb.view(batch_size, *output_shape) if bound_lower else None
    ub = ub.view(batch_size, *output_shape) if bound_upper else None

    if verbose:
        logger.debug('')

    if self.return_A:
        return lb, ub, self.A_dict
    else:
        return lb, ub


def get_unstable_size(unstable_idx):
    if isinstance(unstable_idx, tuple):
        return unstable_idx[0].numel()
    else:
        return unstable_idx.numel()


def check_optimized_variable_sparsity(self: 'BoundedModule', node):
    alpha_sparsity = None  # unknown, optimizable variables are not created for this node.
    for relu in self.relus:
        # FIXME: this hardcoded for ReLUs. Need to support other optimized nonlinear functions.
        # alpha_lookup_idx is only created for sparse-spec alphas.
        if relu.alpha_lookup_idx is not None and node.name in relu.alpha_lookup_idx:
            if relu.alpha_lookup_idx[node.name] is not None:
                # This node was created with sparse alpha
                alpha_sparsity = True
            elif self.bound_opts['optimize_bound_args']['use_shared_alpha']:
                # Shared alpha, the spec dimension is 1, and sparsity can be supported.
                alpha_sparsity = True
            else:
                alpha_sparsity = False
            break
    return alpha_sparsity


def get_sparse_C(self: 'BoundedModule', node, sparse_intermediate_bounds=True,
                 ref_intermediate_lb=None, ref_intermediate_ub=None):
    sparse_conv_intermediate_bounds = self.bound_opts.get('sparse_conv_intermediate_bounds', False)
    minimum_sparsity = self.bound_opts.get('minimum_sparsity', 0.9)
    crown_batch_size = self.bound_opts.get('crown_batch_size', 1e9)
    dim = int(prod(node.output_shape[1:]))
    batch_size = self.batch_size

    reduced_dim = False  # Only partial neurons (unstable neurons) are bounded.
    unstable_idx = None
    unstable_size = np.inf
    newC = None

    alpha_is_sparse = self.check_optimized_variable_sparsity(node)

    # NOTE: batched CROWN is so far only supported for some of the cases below

    # FIXME: C matrix shape incorrect for BoundParams.
    if (isinstance(node, BoundLinear) or isinstance(node, BoundMatMul)) and int(
            os.environ.get('AUTOLIRPA_USE_FULL_C', 0)) == 0:
        if sparse_intermediate_bounds:
            # If we are doing bound refinement and reference bounds are given,
            # we only refine unstable neurons.
            # Also, if we are checking against LP solver we will refine all
            # neurons and do not use this optimization.
            # For each batch element, we find the unstable neurons.
            unstable_idx, unstable_size = self.get_unstable_locations(
                ref_intermediate_lb, ref_intermediate_ub)
            if unstable_size == 0:
                # Do nothing, no bounds will be computed.
                reduced_dim = True
                unstable_idx = []
            elif unstable_size > crown_batch_size:
                # Create C in batched CROWN
                newC = 'OneHot'
                reduced_dim = True
            elif ((0 < unstable_size <= minimum_sparsity * dim
                    and alpha_is_sparse is None) or alpha_is_sparse):
                # When we already have sparse alpha for this layer, we always
                # use sparse C. Otherwise we determine it by sparsity.
                # Create an abstract C matrix, the unstable_idx are the non-zero
                # elements in specifications for all batches.
                newC = OneHotC(
                    [batch_size, unstable_size, *node.output_shape[1:]],
                    self.device, unstable_idx, None)
                reduced_dim = True
            else:
                unstable_idx = None
                del ref_intermediate_lb, ref_intermediate_ub
        if not reduced_dim:
            if dim > crown_batch_size:
                newC = 'eye'
            else:
                newC = eyeC([batch_size, dim, *node.output_shape[1:]], self.device)
    elif node.patches_start and node.mode == "patches":
        if sparse_intermediate_bounds:
            unstable_idx, unstable_size = self.get_unstable_locations(
                ref_intermediate_lb, ref_intermediate_ub, conv=True)
            if unstable_size == 0:
                # Do nothing, no bounds will be computed.
                reduced_dim = True
                unstable_idx = []
            elif unstable_size > crown_batch_size:
                # Create C in batched CROWN
                newC = 'Patches'
                reduced_dim = True
            # We sum over the channel direction, so need to multiply that.
            elif (sparse_conv_intermediate_bounds
                  and unstable_size <= minimum_sparsity * dim
                  and alpha_is_sparse is None) or alpha_is_sparse:
                # When we already have sparse alpha for this layer, we always
                # use sparse C. Otherwise we determine it by sparsity.
                # Create an abstract C matrix, the unstable_idx are the non-zero
                # elements in specifications for all batches.
                # The shape of patches is [unstable_size, batch, C, H, W].
                newC = Patches(
                    shape=[unstable_size, batch_size, *node.output_shape[1:-2],
                           1, 1],
                    identity=1, unstable_idx=unstable_idx,
                    output_shape=[batch_size, *node.output_shape[1:]])
                reduced_dim = True
            else:
                unstable_idx = None
                del ref_intermediate_lb, ref_intermediate_ub
        # Here we create an Identity Patches object
        if not reduced_dim:
            newC = Patches(
                None, 1, 0, [node.output_shape[1], batch_size, *node.output_shape[2:],
                *node.output_shape[1:-2], 1, 1], 1,
                output_shape=[batch_size, *node.output_shape[1:]])
    elif (isinstance(node, (BoundAdd, BoundSub)) and node.mode == "patches"
        and len(node.output_shape) >= 4):
        # FIXME: BoundAdd does not always have patches. Need to use a better way
        # to determine patches mode.
        # FIXME: We should not hardcode BoundAdd here!
        if sparse_intermediate_bounds:
            if crown_batch_size < 1e9:
                warnings.warn('Batched CROWN is not supported in this case')
            unstable_idx, unstable_size = self.get_unstable_locations(
                ref_intermediate_lb, ref_intermediate_ub, conv=True)
            if unstable_size == 0:
                # Do nothing, no bounds will be computed.
                reduced_dim = True
                unstable_idx = []
            elif (sparse_conv_intermediate_bounds
                  and unstable_size <= minimum_sparsity * dim
                  and alpha_is_sparse is None) or alpha_is_sparse:
                # When we already have sparse alpha for this layer, we always
                # use sparse C. Otherwise we determine it by sparsity.
                num_channel = node.output_shape[-3]
                # Identity patch size: (ouc_c, 1, 1, 1, out_c, 1, 1).
                patches = (
                    torch.eye(num_channel, device=self.device,
                    dtype=list(self.parameters())[0].dtype)).view(
                        num_channel, 1, 1, 1, num_channel, 1, 1)
                # Expand to (out_c, 1, unstable_size, out_c, 1, 1).
                patches = patches.expand(-1, 1, node.output_shape[-2],
                                         node.output_shape[-1], -1, 1, 1)
                patches = patches[unstable_idx[0], :,
                                  unstable_idx[1], unstable_idx[2]]
                # Expand with the batch dimension. Final shape
                # (unstable_size, batch_size, out_c, 1, 1).
                patches = patches.expand(-1, batch_size, -1, -1, -1)
                newC = Patches(
                    patches, 1, 0, patches.shape, unstable_idx=unstable_idx,
                    output_shape=[batch_size, *node.output_shape[1:]])
                reduced_dim = True
            else:
                unstable_idx = None
                del ref_intermediate_lb, ref_intermediate_ub
        if not reduced_dim:
            num_channel = node.output_shape[-3]
            # Identity patch size: (ouc_c, 1, 1, 1, out_c, 1, 1).
            patches = (
                torch.eye(num_channel, device=self.device,
                dtype=list(self.parameters())[0].dtype)).view(
                    num_channel, 1, 1, 1, num_channel, 1, 1)
            # Expand to (out_c, batch, out_h, out_w, out_c, 1, 1).
            patches = patches.expand(-1, batch_size, node.output_shape[-2],
                                     node.output_shape[-1], -1, 1, 1)
            newC = Patches(patches, 1, 0, patches.shape, output_shape=[
                batch_size, *node.output_shape[1:]])
    else:
        if sparse_intermediate_bounds:
            unstable_idx, unstable_size = self.get_unstable_locations(
                ref_intermediate_lb, ref_intermediate_ub)
            if unstable_size == 0:
                # Do nothing, no bounds will be computed.
                reduced_dim = True
                unstable_idx = []
            elif unstable_size > crown_batch_size:
                # Create in C in batched CROWN
                newC = 'eye'
                reduced_dim = True
            elif (unstable_size <= minimum_sparsity * dim
                  and alpha_is_sparse is None) or alpha_is_sparse:
                newC = torch.zeros([1, unstable_size, dim], device=self.device)
                # Fill the corresponding elements to 1.0
                newC[0, torch.arange(unstable_size), unstable_idx] = 1.0
                newC = newC.expand(batch_size, -1, -1).view(
                    batch_size, unstable_size, *node.output_shape[1:])
                reduced_dim = True
            else:
                unstable_idx = None
                del ref_intermediate_lb, ref_intermediate_ub
        if not reduced_dim:
            if dim > 1000 and self.verbose:
                warnings.warn(
                    f"Creating an identity matrix with size {dim}x{dim} for node {node}. "
                    "This may indicate poor performance for bound computation. "
                    "If you see this message on a small network please submit "
                    "a bug report.", stacklevel=2)
            if dim > crown_batch_size:
                newC = 'eye'
            else:
                newC = torch.eye(dim, device=self.device).unsqueeze(0).expand(
                    batch_size, -1, -1
                ).view(batch_size, dim, *node.output_shape[1:])

    return newC, reduced_dim, unstable_idx, unstable_size


def restore_sparse_bounds(self: 'BoundedModule', node, unstable_idx,
                          unstable_size, ref_intermediate_lb,
                          ref_intermediate_ub, new_lower=None, new_upper=None):
    batch_size = self.batch_size
    if unstable_size == 0:
        # No unstable neurons. Skip the update.
        node.lower = ref_intermediate_lb.detach().clone()
        node.upper = ref_intermediate_ub.detach().clone()
    else:
        if new_lower is None:
            new_lower = node.lower
        if new_upper is None:
            new_upper = node.upper
        # If we only calculated unstable neurons, we need to scatter the results back based on reference bounds.
        if isinstance(unstable_idx, tuple):
            lower = ref_intermediate_lb.detach().clone()
            upper = ref_intermediate_ub.detach().clone()
            # Conv layer with patches, the unstable_idx is a 3-element tuple for 3 indices (C, H,W) of unstable neurons.
            if len(unstable_idx) == 3:
                lower[:, unstable_idx[0], unstable_idx[1], unstable_idx[2]] = new_lower
                upper[:, unstable_idx[0], unstable_idx[1], unstable_idx[2]] = new_upper
            elif len(unstable_idx) == 4:
                lower[:, unstable_idx[0], unstable_idx[1], unstable_idx[2], unstable_idx[3]] = new_lower
                upper[:, unstable_idx[0], unstable_idx[1], unstable_idx[2], unstable_idx[3]] = new_upper
        else:
            # Other layers.
            lower = ref_intermediate_lb.detach().clone().view(batch_size, -1)
            upper = ref_intermediate_ub.detach().clone().view(batch_size, -1)
            lower[:, unstable_idx] = new_lower.view(batch_size, -1)
            upper[:, unstable_idx] = new_upper.view(batch_size, -1)
        node.lower = lower.view(batch_size, *node.output_shape[1:])
        node.upper = upper.view(batch_size, *node.output_shape[1:])


def get_degrees(node_start):
    if not isinstance(node_start, list):
        node_start = [node_start]
    degrees = {}
    added = {}
    queue = deque()
    for node in node_start:
        queue.append(node)
        added[node.name] = True
    while len(queue) > 0:
        l = queue.popleft()
        for l_pre in l.inputs:
            degrees[l_pre.name] = degrees.get(l_pre.name, 0) + 1
            if not added.get(l_pre.name, False):
                queue.append(l_pre)
                added[l_pre.name] = True
    return degrees


def _preprocess_C(self: 'BoundedModule', C, node):
    if isinstance(C, Patches):
        if C.unstable_idx is None:
            # Patches have size (out_c, batch, out_h, out_w, c, h, w).
            if len(C.shape) == 7:
                out_c, batch_size, out_h, out_w = C.shape[:4]
                output_dim = out_c * out_h * out_w
            else:
                out_dim, batch_size, out_c, out_h, out_w = C.shape[:5]
                output_dim = out_dim * out_c * out_h * out_w
        else:
            # Patches have size (unstable_size, batch, c, h, w).
            output_dim, batch_size = C.shape[:2]
    else:
        batch_size, output_dim = C.shape[:2]

    # The C matrix specified by the user has shape (batch, spec)
    # but internally we have (spec, batch) format.
    if not isinstance(C, (eyeC, Patches, OneHotC)):
        C = C.transpose(0, 1).reshape(
            output_dim, batch_size, *node.output_shape[1:])
    elif isinstance(C, eyeC):
        C = C._replace(shape=(C.shape[1], C.shape[0], *C.shape[2:]))
    elif isinstance(C, OneHotC):
        C = C._replace(
            shape=(C.shape[1], C.shape[0], *C.shape[2:]),
            index=C.index.transpose(0,-1),
            coeffs=None if C.coeffs is None else C.coeffs.transpose(0,-1))

    if isinstance(C, Patches) and C.unstable_idx is not None:
        # Sparse patches; the output shape is (unstable_size, ).
        output_shape = [C.shape[0]]
    elif prod(node.output_shape[1:]) != output_dim and not isinstance(C, Patches):
        # For the output node, the shape of the bound follows C
        # instead of the original output shape
        #
        # TODO Maybe don't set node.lower and node.upper in this case?
        # Currently some codes still depend on node.lower and node.upper
        output_shape = [-1]
    else:
        # Generally, the shape of the bounds match the output shape of the node
        output_shape = node.output_shape[1:]

    return C, batch_size, output_dim, output_shape


def concretize(self, batch_size, output_dim, lb, ub=None,
               bound_lower=True, bound_upper=True,
               average_A=False, node_start=None):
    roots = self.roots()
    for i in range(len(roots)):
        if roots[i].lA is None and roots[i].uA is None: continue
        if average_A and isinstance(roots[i], BoundParams):
            lA = roots[i].lA.mean(
                node_start.batch_dim + 1, keepdim=True
            ).expand(roots[i].lA.shape) if bound_lower else None
            uA = roots[i].uA.mean(
                node_start.batch_dim + 1, keepdim=True
            ).expand(roots[i].uA.shape) if bound_upper else None
        else:
            lA, uA = roots[i].lA, roots[i].uA
        if not isinstance(roots[i].lA, eyeC) and not isinstance(roots[i].lA, Patches):
            lA = roots[i].lA.reshape(output_dim, batch_size, -1).transpose(0, 1) if bound_lower else None
        if not isinstance(roots[i].uA, eyeC) and not isinstance(roots[i].uA, Patches):
            uA = roots[i].uA.reshape(output_dim, batch_size, -1).transpose(0, 1) if bound_upper else None
        if hasattr(roots[i], 'perturbation') and roots[i].perturbation is not None:
            if isinstance(roots[i], BoundParams):
                # add batch_size dim for weights node
                lb = lb + roots[i].perturbation.concretize(
                    roots[i].center.unsqueeze(0), lA,
                    sign=-1, aux=roots[i].aux) if bound_lower else None
                ub = ub + roots[i].perturbation.concretize(
                    roots[i].center.unsqueeze(0), uA,
                    sign=+1, aux=roots[i].aux) if bound_upper else None
            else:
                lb = lb + roots[i].perturbation.concretize(
                    roots[i].center, lA, sign=-1, aux=roots[i].aux) if bound_lower else None
                ub = ub + roots[i].perturbation.concretize(
                    roots[i].center, uA, sign=+1, aux=roots[i].aux) if bound_upper else None
        else:
            fv = roots[i].forward_value
            if type(roots[i]) == BoundInput:
                # Input node with a batch dimension
                batch_size_ = batch_size
            else:
                # Parameter node without a batch dimension
                batch_size_ = 1

            def _add_constant(A, b):
                if isinstance(A, eyeC):
                    b = b + fv.view(batch_size_, -1)
                elif isinstance(A, Patches):
                    b = b + A.matmul(fv, input_shape=roots[0].center.shape)
                elif type(roots[i]) == BoundInput:
                    b = b + A.matmul(fv.view(batch_size_, -1, 1)).squeeze(-1)
                else:
                    b = b + A.matmul(fv.view(-1, 1)).squeeze(-1)
                return b

            lb = _add_constant(lA, lb) if bound_lower else None
            ub = _add_constant(uA, ub) if bound_upper else None

    return lb, ub


def addA(A1, A2):
    """ Add two A (each of them is either Tensor or Patches) """
    if type(A1) == type(A2):
        return A1 + A2
    elif type(A1) == Patches:
        return A1 + A2
    elif type(A2) == Patches:
        return A2 + A1
    else:
        raise NotImplementedError(f'Unsupported types for A1 ({type(A1)}) and A2 ({type(A2)}')


def add_bound(node, node_pre, lA=None, uA=None):
    """Propagate lA and uA to a preceding node."""
    if lA is not None:
        if node_pre.lA is None:
            # First A added to this node.
            node_pre.zero_lA_mtx = node.zero_backward_coeffs_l
            node_pre.lA = lA
        else:
            node_pre.zero_lA_mtx = node_pre.zero_lA_mtx and node.zero_backward_coeffs_l
            new_node_lA = addA(node_pre.lA, lA)
            node_pre.lA = new_node_lA
    if uA is not None:
        if node_pre.uA is None:
            # First A added to this node.
            node_pre.zero_uA_mtx = node_pre.zero_backward_coeffs_u
            node_pre.uA = uA
        else:
            node_pre.zero_uA_mtx = node_pre.zero_uA_mtx and node.zero_backward_coeffs_u
            node_pre.uA = addA(node_pre.uA, uA)


def add_constant_node(lb, ub, node):
    new_lb = node.get_bias(node.lA, node.forward_value)
    new_ub = node.get_bias(node.uA, node.forward_value)
    if isinstance(lb, Tensor) and isinstance(new_lb, Tensor) and lb.ndim > 0 and lb.ndim != new_lb.ndim:
        new_lb = new_lb.reshape(lb.shape)
    if isinstance(ub, Tensor) and isinstance(new_ub, Tensor) and ub.ndim > 0 and ub.ndim != new_ub.ndim:
        new_ub = new_ub.reshape(ub.shape)
    lb = lb + new_lb # FIXME (09/16): shape for the bias of BoundConstant.
    ub = ub + new_ub
    return lb, ub


def save_A_record(node, A_record, A_dict, roots, needed_A_dict, lb, ub, unstable_idx):
    root_A_record = {}
    for i in range(len(roots)):
        if roots[i].lA is None and roots[i].uA is None: continue
        if roots[i].name in needed_A_dict:
            if roots[i].lA is not None:
                if isinstance(roots[i].lA, Patches):
                    _lA = roots[i].lA
                else:
                    _lA = roots[i].lA.transpose(0, 1).detach()
            else:
                _lA = None

            if roots[i].uA is not None:
                if isinstance(roots[i].uA, Patches):
                    _uA = roots[i].uA
                else:
                    _uA = roots[i].uA.transpose(0, 1).detach()
            else:
                _uA = None
            root_A_record.update({roots[i].name: {
                "lA": _lA,
                "uA": _uA,
                # When not used, lb or ub is tensor(0). They have been transposed above.
                "lbias": lb.detach() if lb.ndim > 1 else None,
                "ubias": ub.detach() if ub.ndim > 1 else None,
                "unstable_idx": unstable_idx
            }})
    root_A_record.update(A_record)  # merge to existing A_record
    A_dict.update({node.name: root_A_record})


def select_unstable_idx(ref_intermediate_lb, ref_intermediate_ub, unstable_locs, max_crown_size):
    """When there are too many unstable neurons, only bound those
    with the loosest reference bounds."""
    gap = (
        ref_intermediate_ub[:, unstable_locs]
        - ref_intermediate_lb[:, unstable_locs]).sum(dim=0)
    indices = torch.argsort(gap, descending=True)
    indices_selected = indices[:max_crown_size]
    indices_selected, _ = torch.sort(indices_selected)
    print(f'{len(indices_selected)}/{len(indices)} unstable neurons selected for CROWN')
    return indices_selected


def get_unstable_locations(self: 'BoundedModule', ref_intermediate_lb,
                           ref_intermediate_ub, conv=False, channel_only=False):
    # FIXME (2023): This function should be a member class of the Bound object, since the
    # definition of unstable neurons depends on the activation function.
    max_crown_size = self.bound_opts.get('max_crown_size', int(1e9))
    # For conv layer we only check the case where all neurons are active/inactive.
    unstable_masks = torch.logical_and(ref_intermediate_lb < 0, ref_intermediate_ub > 0)
    # For simplicity, merge unstable locations for all elements in this batch. TODO: use individual unstable mask.
    # It has shape (H, W) indicating if a neuron is unstable/stable.
    # TODO: so far we merge over the batch dimension to allow easier implementation.
    if channel_only:
        # Only keep channels with unstable neurons. Used for initializing alpha.
        unstable_locs = unstable_masks.sum(dim=(0,2,3)).bool()
        # Shape is consistent with linear layers: a list of unstable neuron channels (no batch dim).
        unstable_idx = unstable_locs.nonzero().squeeze(1)
    else:
        if not conv and unstable_masks.ndim > 2:
            # Flatten the conv layer shape.
            unstable_masks = unstable_masks.reshape(unstable_masks.size(0), -1)
            ref_intermediate_lb = ref_intermediate_lb.reshape(ref_intermediate_lb.size(0), -1)
            ref_intermediate_ub = ref_intermediate_ub.reshape(ref_intermediate_ub.size(0), -1)
        unstable_locs = unstable_masks.sum(dim=0).bool()
        if conv:
            # Now converting it to indices for these unstable nuerons.
            # These are locations (i,j) of unstable neurons.
            unstable_idx = unstable_locs.nonzero(as_tuple=True)
        else:
            unstable_idx = unstable_locs.nonzero().squeeze(1)

    unstable_size = get_unstable_size(unstable_idx)
    if unstable_size > max_crown_size:
        indices_seleted = select_unstable_idx(
            ref_intermediate_lb, ref_intermediate_ub, unstable_locs, max_crown_size)
        if isinstance(unstable_idx, tuple):
            unstable_idx = tuple(u[indices_seleted] for u in unstable_idx)
        else:
            unstable_idx = unstable_idx[indices_seleted]
    unstable_size = get_unstable_size(unstable_idx)

    return unstable_idx, unstable_size


def get_alpha_crown_start_nodes(
        self: 'BoundedModule',
        node,
        c=None,
        share_alphas=False,
        final_node_name=None,
        backward_from_node: Bound = None,
    ):
    """
    Given a layer "node", return a list of following nodes after this node whose bounds
    will propagate through this node. Each element in the list is a tuple with 3 elements:
    (following_node_name, following_node_shape, unstable_idx)
    """
    # When use_full_conv_alpha is True, conv layers do not share alpha.
    sparse_intermediate_bounds = self.bound_opts.get('sparse_intermediate_bounds', False)
    use_full_conv_alpha_thresh = self.bound_opts.get('use_full_conv_alpha_thresh', 512)

    start_nodes = []
    # In most cases, backward_from_node == node
    # Only if output constraints are used, will they differ: the node that should be
    # bounded (node) needs alphas for *all* layers, not just those behind it.
    # In this case, backward_from_node will be the input node
    assert backward_from_node == node

    for nj in self.backward_from[backward_from_node.name]:  # Pre-activation layers.
        unstable_idx = None
        use_sparse_conv = None  # Whether a sparse-spec alpha is used for a conv output node. None for non-conv output node.
        use_full_conv_alpha = self.bound_opts.get('use_full_conv_alpha', False)

        # Find the indices of unstable neuron, used for create sparse-feature alpha.
        if (sparse_intermediate_bounds
                and isinstance(node, BoundOptimizableActivation)
                and nj.name != final_node_name and not share_alphas):
            # Create sparse optimization variables for intermediate neurons.
            # These are called "sparse-spec" alpha because we only create alpha only for
            # the intermediate of final output nodes whose bounds are needed.
            # "sparse-spec" alpha makes sense only for piece-wise linear functions.
            # For other intermediate nodes, there is no "unstable" or "stable" neuron.
            # FIXME: whether an layer has unstable/stable neurons should be in Bound obj.
            # FIXME: get_unstable_locations should be a member class of ReLU.
            if len(nj.output_name) == 1 and isinstance(self[nj.output_name[0]], (BoundRelu, BoundSignMerge, BoundMaxPool)):
                if ((isinstance(nj, (BoundLinear, BoundMatMul)))
                        and int(os.environ.get('AUTOLIRPA_USE_FULL_C', 0)) == 0):
                    # unstable_idx has shape [neuron_size_of_nj]. Batch dimension is reduced.
                    unstable_idx, _ = self.get_unstable_locations(nj.lower, nj.upper)
                elif isinstance(nj, (BoundConv, BoundAdd, BoundSub, BoundBatchNormalization)) and nj.mode == 'patches':
                    if nj.name in node.patch_size:
                        # unstable_idx has shape [channel_size_of_nj]. Batch and spatial dimensions are reduced.
                        unstable_idx, _ = self.get_unstable_locations(
                            nj.lower, nj.upper, channel_only=not use_full_conv_alpha, conv=True)
                        use_sparse_conv = False  # alpha is shared among channels. Sparse-spec alpha in hw dimension not used.
                        if use_full_conv_alpha and unstable_idx[0].size(0) > use_full_conv_alpha_thresh:
                            # Too many unstable neurons. Using shared alpha per channel.
                            unstable_idx, _ = self.get_unstable_locations(
                                nj.lower, nj.upper, channel_only=True, conv=True)
                            use_full_conv_alpha = False
                    else:
                        # Matrix mode for conv layers. Although the bound propagation started with patches mode,
                        # when A matrix is propagated to this layer, it might become a dense matrix since patches
                        # can be come very large after many layers. In this case,
                        # unstable_idx has shape [c_out * h_out * w_out]. Batch dimension is reduced.
                        unstable_idx, _ = self.get_unstable_locations(nj.lower, nj.upper)
                        use_sparse_conv = True  # alpha is not shared among channels, and is sparse in spec dimension.
            else:
                # FIXME: we should not check for fixed names here. Need to enable patches mode more generally.
                if isinstance(nj, (BoundConv, BoundAdd, BoundSub, BoundBatchNormalization)) and nj.mode == 'patches':
                    use_sparse_conv = False  # Sparse-spec alpha can never be used, because it is not a ReLU activation.

        if nj.name == final_node_name:
            # Final layer, always the number of specs as the shape.
            size_final = self[final_node_name].output_shape[-1] if c is None else c.size(1)
            # The 4-th element indicates that this start node is the final node,
            # which may be utilized by operators that do not know the name of
            # the final node.
            start_nodes.append((final_node_name, size_final, None, True))
            continue

        if share_alphas:
            # all intermediate neurons from the same layer share the same set of alphas.
            output_shape = 1
        elif isinstance(node, BoundOptimizableActivation) and node.patch_size and nj.name in node.patch_size:
            # Patches mode. Use output channel size as the spec size. This still shares some alpha, but better than no sharing.
            if use_full_conv_alpha:
                # alphas not shared among channels, so the spec dim shape is c,h,w
                # The patch size is [out_ch, batch, out_h, out_w, in_ch, H, W]. We use out_ch as the output shape.
                output_shape = node.patch_size[nj.name][0], node.patch_size[nj.name][2], node.patch_size[nj.name][3]
            else:
                # The spec dim is c only, and is shared among h, w.
                output_shape = node.patch_size[nj.name][0]
            assert not sparse_intermediate_bounds or use_sparse_conv is False  # Double check our assumption holds. If this fails, then we created wrong shapes for alpha.
        else:
            # Output is linear layer (use_sparse_conv = None), or patch converted to matrix (use_sparse_conv = True).
            assert not sparse_intermediate_bounds or use_sparse_conv is not False  # Double check our assumption holds. If this fails, then we created wrong shapes for alpha.
            output_shape = nj.lower.shape[1:]  # FIXME: for non-relu activations it's still expecting a prod.
        start_nodes.append((nj.name, output_shape, unstable_idx, False))
    return start_nodes


def merge_A(batch_A, ret_A):
    for key0 in batch_A:
        if key0 not in ret_A: 
            ret_A[key0] = {}
        for key1 in batch_A[key0]:
            value = batch_A[key0][key1]
            if key1 not in ret_A[key0]:
                # create:
                ret_A[key0].update({
                    key1: {
                        "lA": value["lA"],
                        "uA": value["uA"],
                        "lbias": value["lbias"],
                        "ubias": value["ubias"],
                        "unstable_idx": value["unstable_idx"]
                    }
                })
            # elif key0 == node.name:
            else:
                # merge:
                # the batch splitting only happens for current node, i.e.,
                # for other nodes the returned lA should be the same across different batches
                # so no need to repeatly merge them
                exist = ret_A[key0][key1]

                if exist["unstable_idx"] is not None:
                    if isinstance(exist["unstable_idx"], torch.Tensor):
                        merged_unstable = torch.cat([
                            exist["unstable_idx"],
                            value['unstable_idx']], dim=0)
                    elif isinstance(exist["unstable_idx"], tuple):
                        if exist["unstable_idx"]:
                            merged_unstable = tuple([
                                torch.cat([exist["unstable_idx"][idx],
                                           value['unstable_idx'][idx]], dim=0)
                                for idx in range(len(exist['unstable_idx']))]
                            )
                        else:
                            merged_unstable = None
                    else:
                        raise NotImplementedError(
                            f'Unsupported type {type(exist["unstable_idx"])}')
                else:
                    merged_unstable = None
                merge_dict = {"unstable_idx": merged_unstable}
                for name in ["lA", "uA"]:
                    if exist[name] is not None:
                        if isinstance(exist[name], torch.Tensor):
                            # for matrix the spec dim is 1
                            merge_dict[name] = torch.cat([exist[name], value[name]], dim=1)
                        else:
                            assert isinstance(exist[name], Patches)
                            # for patches the spec dim`is 0
                            merge_dict[name] = exist[name].create_similar(
                                torch.cat([exist[name].patches, value[name].patches], dim=0),
                                unstable_idx=merged_unstable
                            )
                    else:
                        merge_dict[name] = None
                for name in ["lbias", "ubias"]:
                    if exist[name] is not None:
                        # for bias the spec dim in 1
                        merge_dict[name] = torch.cat([exist[name], value[name]], dim=1)
                    else:
                        merge_dict[name] = None
                ret_A[key0][key1] = merge_dict
    return ret_A



import logging
import time
import torch
import torch.nn as nn
import os
import sys
from collections import defaultdict, namedtuple
from functools import reduce
import operator
import warnings
from typing import Tuple
from .patches import Patches

logging.basicConfig(
    format='%(levelname)-8s %(asctime)-12s %(message)s',
    datefmt='%H:%M:%S',
    stream=sys.stdout
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG if os.environ.get('AUTOLIRPA_DEBUG', 0) else logging.INFO)

warnings.simplefilter("once")

# Special identity matrix. Avoid extra computation of identity matrix multiplication in various places.
eyeC = namedtuple('eyeC', 'shape device')
OneHotC = namedtuple('OneHotC', 'shape device index coeffs')

def onehotc_to_dense(one_hot_c: OneHotC, dtype: torch.dtype) -> torch.Tensor:
    shape = one_hot_c.shape  # [spec, batch, C, H, W]
    dim = int(prod(shape[2:]))
    dense = torch.zeros(
        size=(shape[0], shape[1], dim), device=one_hot_c.device, dtype=dtype)
    # one_hot_c.index has size (spec, batch), its values are the index of the one-hot non-zero elements in A.
    # one_hot_c.coeffs is the value of the non-zero element.
    dense = torch.scatter(
        dense, dim=2, index=one_hot_c.index.unsqueeze(-1),
        src=one_hot_c.coeffs.unsqueeze(-1))
    dense = dense.view(shape[0], shape[1], *shape[2:])
    return dense

# Benchmarking mode disable some expensive assertions.
Benchmarking = True

reduction_sum = lambda x: x.sum(1, keepdim=True)
reduction_mean = lambda x: x.mean(1, keepdim=True)
reduction_max = lambda x: x.max(1, keepdim=True).values
reduction_min = lambda x: x.min(1, keepdim=True).values

MIN_HALF_FP = 5e-8  # 2**-24, which is the smallest value that float16 can be represented


def reduction_str2func(reduction_func):
    if type(reduction_func) == str:
        if reduction_func == 'min':
            return reduction_min
        elif reduction_func == 'max':
            return reduction_max
        elif reduction_func == 'sum':
            return reduction_sum
        elif reduction_func == 'mean':
            return reduction_mean
        else:
            raise NotImplementedError(f'Unknown reduction_func {reduction_func}')
    else:
        return reduction_func

def stop_criterion_placeholder(threshold=0):
    return lambda x: RuntimeError("BUG: bound optimization stop criterion not specified.")

def stop_criterion_min(threshold=0):
    return lambda x: (x.min(1, keepdim=True).values > threshold)

def stop_criterion_all(threshold=0):
    # The dimension of x should be (batch, spec). The spec dimension
    # This was used in the incomplete verifier, where the spec dimension can
    # present statements in an OR clause.
    return lambda x: (x > threshold).all(dim=1, keepdim=True)

def stop_criterion_max(threshold=0):
    return lambda x: (x.max(1, keepdim=True).values > threshold)

def stop_criterion_batch(threshold=0):
    # may unexpected broadcast, pay attention to the shape of threshold
    # x shape: batch, number_bounds; threshold shape: batch, number_bounds
    return lambda x: (x > threshold)

def stop_criterion_batch_any(threshold=0):
    """If any spec >= rhs, then this sample can be stopped;
       if all samples can be stopped, stop = True, o.w., False.
    """
    # may unexpected broadcast, pay attention to the shape of threshold
    # x shape: batch, number_bounds; threshold shape: batch, number_bounds
    return lambda x: (x > threshold).any(dim=1, keepdim=True)

def stop_criterion_batch_topk(threshold=0, k=1314):
    # x shape: batch, number_bounds; threshold shape: batch, number_bounds
    return lambda x: (torch.kthvalue(x, k, dim=-1, keepdim=True).values > threshold).any(dim=1)

def multi_spec_keep_func_all(x):
    return torch.all(x, dim=-1)



class MultiAverageMeter(object):
    """Computes and stores the average and current value for multiple metrics"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.sum_meter = defaultdict(float)
        self.lasts = defaultdict(float)
        self.counts_meter = defaultdict(int)
        self.batch_size = 1

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size

    def update(self, key, val, n=None):
        if val is None:
            return
        if n is None:
            n = self.batch_size
        if isinstance(val, torch.Tensor):
            val = val.item()
        self.lasts[key] = val
        self.sum_meter[key] += val * n
        self.counts_meter[key] += n

    def last(self, key):
        return self.lasts[key]

    def avg(self, key):
        if self.counts_meter[key] == 0:
            return 0.0
        else:
            return self.sum_meter[key] / self.counts_meter[key]

    def __repr__(self):
        s = ""
        for k in self.sum_meter:
            s += "{}={:.4f} ".format(k, self.avg(k))
        return s.strip()


class MultiTimer(object):
    """Count the time for each part of training."""
    def __init__(self):
        self.reset()
    def reset(self):
        self.timer_starts = defaultdict(float)
        self.timer_total = defaultdict(float)
    def start(self, key):
        if self.timer_starts[key] != 0:
            raise RuntimeError("start() is called more than once")
        self.timer_starts[key] = time.time()
    def stop(self, key):
        if key not in self.timer_starts:
            raise RuntimeError("Key does not exist; please call start() before stop()")
        self.timer_total[key] += time.time() - self.timer_starts[key]
        self.timer_starts[key] = 0
    def total(self, key):
        return self.timer_total[key]
    def __repr__(self):
        s = ""
        for k in self.timer_total:
            s += "{}_time={:.3f} ".format(k, self.timer_total[k])
        return s.strip()


class Flatten(nn.Flatten):
    """Legacy Flatten class.

    It was previously created when nn.Flatten was not supported. Simply use
    nn.Flatten in the future."""
    pass


class Unflatten(nn.Module):
    def __init__(self, wh):
        super().__init__()
        self.wh = wh # width and height of the feature maps
    def forward(self, x):
        return x.view(x.size(0), -1, self.wh, self.wh)


class Max(nn.Module):

    def __init__(self):
        super(Max, self).__init__()

    def forward(self, x, y):
        return torch.max(x, y)


class Min(nn.Module):

    def __init__(self):
        super(Min, self).__init__()

    def forward(self, x, y):
        return torch.min(x, y)


def scale_gradients(optimizer, gradient_accumulation_steps, grad_clip=None):
    parameters = []
    for param_group in optimizer.param_groups:
        for param in param_group['params']:
            parameters.append(param)
            if param.grad is not None:
                param.grad.data /= gradient_accumulation_steps
    if grad_clip is not None:
        return torch.nn.utils.clip_grad_norm_(parameters, grad_clip)


# unpack tuple, dict, list into one single list
# TODO: not sure if the order matches graph.inputs()
def unpack_inputs(inputs, device=None):
    if isinstance(inputs, dict):
        inputs = list(inputs.values())
    if isinstance(inputs, tuple) or isinstance(inputs, list):
        res = []
        for item in inputs:
            res += unpack_inputs(item, device=device)
        return res
    else:
        if device is not None:
            inputs = inputs.to(device)
        return [inputs]


def isnan(x):
    if isinstance(x, Patches):
        return False
    return torch.isnan(x).any()


def prod(x):
    return reduce(operator.mul, x, 1)


def batched_index_select(input, dim, index):
    # Assuming the input has a batch dimension.
    # index has dimensin [spec, batch].
    if input.ndim == 4:
        # Alphas for fully connected layers, shape [2, spec, batch, neurons]
        index = index.unsqueeze(-1).unsqueeze(0).expand(input.size(0), -1, -1, input.size(3))
    elif input.ndim == 6:
        # Alphas for fully connected layers, shape [2, spec, batch, c, h, w].
        index = index.view(1, index.size(0), index.size(1), *([1] * (input.ndim - 3))).expand(input.size(0), -1, -1, *input.shape[3:])
    elif input.ndim == 3:
        # Weights.
        input = input.expand(index.size(0), -1, -1)
        index = index.unsqueeze(-1).expand(-1, -1, input.size(2))
    elif input.ndim == 2:
        # Bias.
        input = input.expand(index.size(0), -1)
    else:
        raise ValueError
    return torch.gather(input, dim, index)


def get_spec_matrix(X, y, num_classes):
    with torch.no_grad():
        c = (torch.eye(num_classes).type_as(X)[y].unsqueeze(1)
            - torch.eye(num_classes).type_as(X).unsqueeze(0))
        I = (~(y.unsqueeze(1) == torch.arange(num_classes).type_as(y).unsqueeze(0)))
        c = (c[I].view(X.size(0), num_classes - 1, num_classes))
    return c


def unravel_index(
    indices: torch.LongTensor,
    shape: Tuple[int, ...],
) -> torch.LongTensor:
    r"""Converts flat indices into unraveled coordinates in a target shape.

    Args:
        indices: A tensor of (flat) indices, (*, N).
        shape: The targeted shape, (D,).

    Returns:
        The unraveled coordinates, a list with tensors in shape (N, D).

    Code borrowed from:
        https://github.com/pytorch/pytorch/issues/35674
    """

    coord = []

    for dim in reversed(shape):
        coord.append(indices % dim)
        indices = torch.div(indices, dim, rounding_mode='trunc')

    return list(reversed(coord))


def fill_template(out, template):
    if template is None:
        return out.popleft()
    elif isinstance(template, (list, tuple)):
        res = []
        for t in template:
            res.append(fill_template(t))
        return tuple(res) if isinstance(template, tuple) else res
    elif isinstance(template, dict):
        res = {}
        for key in template:
            res[key] = fill_template(template[key])
        return res
    else:
        raise NotImplementedError



import logging
import torch
from util.misc.logger import logger
from attacker.attacker import Attacker, PGDAttacker
from test import extract_instance


def get_activation_shape(name, result):
    def hook(model, input, output):
        result[name] = output.shape
    return hook


def attack(onnx_name, vnnlib_name, timeout, device):
    model, input_shape, objectives = extract_instance(onnx_name, vnnlib_name)
    model.to(device)
    print(model)
    print(f'{input_shape=}')
    
    # Define a list to store the outputs
    outputs = []

    # Define the hook function
    # def hook_fn(module, input, output):
    #     outputs.append(output)
        
    # for layer in model.children():
    #     if isinstance(layer, torch.nn.ReLU):
    #         layer.register_forward_hook(hook_fn)
    
    # atk = PGDAttacker(model, objectives, input_shape, device)
    # is_attacked, adv = atk.run(iterations=10, restarts=5, timeout=timeout)

    atk = Attacker(model, objectives, input_shape, device)
    is_attacked, adv = atk.run(timeout=timeout)
    
    # for i, output in enumerate(outputs):
    #     print(f"Output after ReLU layer {i+1}: {output}")
    # print(adv)
    print('adv dtype=', adv.dtype if adv is not None else None)
    if is_attacked:
        assert adv is not None
        return 'sat'
    
    return 'unknown'
    
    
if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    net_name = 'example/onnx/mnist-net_256x2.onnx'
    vnnlib_name = 'example/vnnlib/prop_1_0.05.vnnlib'

    net_name = 'example/onnx/motivation_example.onnx'
    vnnlib_name = 'example/vnnlib/motivation_example.vnnlib'
    
    preconditions = [
        [],
        [],
    ]
    print(attack(net_name, vnnlib_name, 2.0, 'cpu'))
    


from torch.profiler import profile, record_function, ProfilerActivity
import torchvision.models as models
import torch

from verifier.verifier import Verifier 
from .test_function import extract_instance
from util.misc.logger import logger
from setting import Settings



    
    
if __name__ == "__main__":
    device = 'cuda'
    # model = models.resnet18().to(device)
    # inputs = torch.randn(50, 3, 224, 224).to(device)
    
    net_path = 'example/backup/motivation_example_159.onnx'
    vnnlib_path = 'example/backup/motivation_example_159.vnnlib'
    
    net_path = 'example/onnx/mnist-net_256x2.onnx'
    vnnlib_path = 'example/vnnlib/prop_1_0.03.vnnlib'
    
    logger.setLevel(1)
    # Settings.setup_test()
    Settings.setup(args=None)
    
    model, input_shape, objectives = extract_instance(net_path, vnnlib_path)
    model.to(device)
    verifier = Verifier(
        net=model, 
        input_shape=input_shape, 
        batch=1000,
        device=device,
    )
    
    with profile(activities=[ProfilerActivity.CUDA, ProfilerActivity.CPU], record_shapes=False) as prof:
        with record_function("test"):
            # model(inputs)
            verifier.verify(objectives)
            
    print(prof.key_averages().table(sort_by="cuda_time_total", row_limit=50))


from .base import *
from .linear import *
from .convolution import *
from .pooling import *
from .activation_base import *
from .activations import *
from .nonlinear import *
from .relu import *
from .tanh import *
from .bivariate import *
from .add_sub import *
from .normalization import *
from .shape import *
from .reduce import *
from .rnn import *
from .softmax import *
from .constant import *
from .leaf import *
from .logical import *
from .dropout import *
from .dtype import *
from .trigonometric import *
from .gradient_bounds import *
from .gradient_modules import *
from .solver_utils import grb
from .resize import *
from .jacobian import *



from __future__ import annotations
import warnings
warnings.filterwarnings(action='ignore')
from beartype import beartype
import torch
import copy

from ..heuristic.decision_heuristics import DecisionHeuristic
from ..auto_LiRPA.utils import stop_criterion_batch_any
from ..onnx2pytorch.convert.model import ConvertModel
from ..heuristic.domains_list import DomainsList
from ..abstractor.utils import new_slopes


class InteractiveVerifier:

    "Branch-and-Bound Interactive Verifier"

    # @beartype
    def __init__(self: 'InteractiveVerifier', net: ConvertModel | torch.nn.Module , input_shape: tuple, batch: int = 1000, device: str = 'cpu') -> None:
        self.net = net # pytorch model
        self.input_shape = input_shape
        self.device = device

        # hyper parameters
        self.input_split = False
        self.batch = max(batch, 1)

        self.scorer = DecisionHeuristic(
            input_split=self.input_split,
            decision_topk=-1,
            decision_method='greedy'
        )


    # @beartype
    def _initialize(self: 'InteractiveVerifier', objective, preconditions: dict, reference_bounds: dict | None) -> DomainsList | list:
        # initialization params
        ret = self.abstractor.initialize(objective, reference_bounds=reference_bounds)

        # check verified
        assert len(ret.output_lbs) == len(objective.cs)
        if stop_criterion_batch_any(objective.rhs.to(self.device))(ret.output_lbs.to(self.device)).all():
            return []

        # full slopes uses too much memory
        slopes = ret.slopes if self.input_split else new_slopes(ret.slopes, self.abstractor.net.final_name)

        # remaining domains
        return DomainsList(
            net=self.abstractor.net,
            objective_ids=ret.objective_ids,
            output_lbs=ret.output_lbs,
            input_lowers=ret.input_lowers,
            input_uppers=ret.input_uppers,
            lower_bounds=ret.lower_bounds,
            upper_bounds=ret.upper_bounds,
            lAs=ret.lAs,
            slopes=slopes, # pruned slopes
            histories=copy.deepcopy(ret.histories),
            cs=ret.cs,
            rhs=ret.rhs,
            input_split=self.input_split,
            preconditions=preconditions,
        )


    def decide(self, observation, domain_params):
        decisions = self.decision(self.abstractor, domain_params)
        return (decisions, domain_params), None


    def init(self: 'InteractiveVerifier', objective, preconditions: dict, reference_bounds: dict | None) -> DomainsList | list:
        self._setup_restart(0, objective)
        self.domains_list = self._initialize(
            objective=objective,
            preconditions=preconditions,
            reference_bounds=reference_bounds,
        )
        return len(self.domains_list) == 0


    def get_observation(self, batch):
        pick_ret = self.domains_list.pick_out(batch, self.device)

        obs = self.scorer.get_branching_scores(
            abstractor=self.abstractor,
            domain_params=pick_ret,
        )
        # topk_output_lbs: (topk, batch)
        # topk_decisions: (topk, batch)
        return obs, pick_ret

    def get_rewards(self, pick_ret, reduce_op=torch.max):
        topk_output_lbs, topk_decisions = self.scorer.get_all_branching_rewards(
            abstractor=self.abstractor,
            domain_params=pick_ret,
            reduce_op=reduce_op,
        )
        return topk_output_lbs, topk_decisions

    def step(self, action):
        decisions, pick_ret = action
        abstraction_ret = self.abstractor.forward(decisions, pick_ret)
        self.domains_list.add(abstraction_ret, decisions)
        done = len(self.domains_list) == 0

        info = {
            'worst_bound': self.domains_list.minimum_lowers,
            'visited': self.domains_list.visited,
            'remaining': len(self.domains_list),
        }

        # reward = self.domains_list.minimum_lowers
        # given action (selected neuron)
        # return minimum lowerbound on two branches of that neuron
        reward = abstraction_ret.output_lbs.min(dim=-1).values
        r1, r2 = torch.chunk(reward, 2)
        reward = torch.minimum(r1, r2)

        return reward, done, info

    from .utils import _preprocess, _init_abstractor, _setup_restart



""" Leaf nodes (indepedent nodes in the auto_LiRPA paper).

Including input, parameter, buffer, etc."""

from itertools import chain
from .base import *


class BoundInput(Bound):
    def __init__(self, ori_name, value, perturbation=None, input_index=None):
        super().__init__()
        self.ori_name = ori_name
        self.value = value
        self.perturbation = perturbation
        self.from_input = True
        self.input_index = input_index

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        # Update perturbed property based on the perturbation set.
        if key == "perturbation":
            if self.perturbation is not None:
                self.perturbed = True
            else:
                self.perturbed = False

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        r"""Copies parameters and buffers from :attr:`state_dict` into only
        this module, but not its descendants. This is called on every submodule
        in :meth:`~torch.nn.Module.load_state_dict`. Metadata saved for this
        module in input :attr:`state_dict` is provided as :attr:`local_metadata`.
        For state dicts without metadata, :attr:`local_metadata` is empty.
        Subclasses can achieve class-specific backward compatible loading using
        the version number at `local_metadata.get("version", None)`.

        .. note::
            :attr:`state_dict` is not the same object as the input
            :attr:`state_dict` to :meth:`~torch.nn.Module.load_state_dict`. So
            it can be modified.

        Arguments:
            state_dict (dict): a dict containing parameters and
                persistent buffers.
            prefix (str): the prefix for parameters and buffers used in this
                module
            local_metadata (dict): a dict containing the metadata for this module.
                See
            strict (bool): whether to strictly enforce that the keys in
                :attr:`state_dict` with :attr:`prefix` match the names of
                parameters and buffers in this module
            missing_keys (list of str): if ``strict=True``, add missing keys to
                this list
            unexpected_keys (list of str): if ``strict=True``, add unexpected
                keys to this list
            error_msgs (list of str): error messages should be added to this
                list, and will be reported together in
                :meth:`~torch.nn.Module.load_state_dict`
        """
        for hook in self._load_state_dict_pre_hooks.values():
            hook(state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

        local_name_params = chain(self._parameters.items(), self._buffers.items())
        local_state = {k: v for k, v in local_name_params if v is not None}

        for name, param in local_state.items():
            if len(prefix.split('.')) == 2:
                key = prefix + name
            else:
                # change key to prefix + self.ori_name when calling load_state_dict()
                key = '.'.join(prefix.split('.')[:-2]) + '.' + self.ori_name
            if key in state_dict:
                input_param = state_dict[key]

                # Backward compatibility: loading 1-dim tensor from 0.3.* to version 0.4+
                if param.ndim == 0 and input_param.ndim == 1:
                    input_param = input_param[0]

                if input_param.shape != param.shape:
                    # local shape should match the one in checkpoint
                    error_msgs.append('size mismatch for {}: copying a param with shape {} from checkpoint, '
                                      'the shape in current model is {}.'
                                      .format(key, input_param.shape, param.shape))
                    continue

                try:
                    with torch.no_grad():
                        param.copy_(input_param)
                except Exception as ex:
                    error_msgs.append('While copying the parameter named "{}", '
                                      'whose dimensions in the model are {} and '
                                      'whose dimensions in the checkpoint are {}, '
                                      'an exception occured : {}.'
                                      .format(key, param.size(), input_param.size(), ex.args))
            elif strict:
                missing_keys.append(key)

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        r"""Saves module state to `destination` dictionary, containing a state
        of the module, but not its descendants. This is called on every
        submodule in :meth:`~torch.nn.Module.state_dict`.

        In rare cases, subclasses can achieve class-specific behavior by
        overriding this method with custom logic.

        Arguments:
            destination (dict): a dict where state will be stored
            prefix (str): the prefix for parameters and buffers used in this
                module
        """
        for param in self._parameters.values():
            if param is not None:
                if len(prefix.split('.')) == 2:
                    destination[self.ori_name] = param if keep_vars else param.detach()
                else:
                    # change parameters' name to self.ori_name when calling state_dict()
                    destination[
                        '.'.join(prefix.split('.')[:-2]) + '.' + self.ori_name] = param if keep_vars else param.detach()
        for buf in self._buffers.values():
            if buf is not None:
                if len(prefix.split('.')) == 2:
                    destination[self.ori_name] = buf if keep_vars else buf.detach()
                else:
                    # change buffers' name to self.ori_name when calling state_dict()
                    destination[
                        '.'.join(prefix.split('.')[:-2]) + '.' + self.ori_name] = buf if keep_vars else buf.detach()

    def forward(self):
        return self.value

    def bound_forward(self, dim_in):
        assert 0

    def bound_backward(self, last_lA, last_uA, **kwargs):
        raise ValueError('{} is a BoundInput node and should not be visited here'.format(
            self.name))

    def interval_propagate(self, *v):
        raise ValueError('{} is a BoundInput node and should not be visited here'.format(
            self.name))

class BoundParams(BoundInput):
    def __init__(self, ori_name, value, perturbation=None, options=None):
        super().__init__(ori_name, None, perturbation)
        self.register_parameter('param', value)
        self.from_input = False
        self.initializing = False

    def register_parameter(self, name, param):
        """Override register_parameter() hook to register only needed parameters."""
        if name == 'param':
            return super().register_parameter(name, param)
        else:
            # Just register it as a normal property of class.
            object.__setattr__(self, name, param)

    def init(self, initializing=False):
        self.initializing = initializing

    def forward(self):
        if self.initializing:
            return self.param_init.requires_grad_(self.training)
        else:
            return self.param.requires_grad_(self.training)

class BoundBuffers(BoundInput):
    def __init__(self, ori_name, value, perturbation=None, options=None):
        super().__init__(ori_name, None, perturbation)
        self.register_buffer('buffer', value.clone().detach())
        self.from_input = not options.get('buffers', {}).get('no_batchdim', False)

    def forward(self):
        return self.buffer



from __future__ import annotations
from beartype import beartype
import random
import torch
import copy
import time

from ..util.misc.logger import logger

class SATSolver:

    # @beartype
    def __init__(self: 'SATSolver', clauses: list[list]) -> None:

        self._last_assigned_literals = False
        self.assignment = {}

        # add clauses
        if len(clauses):
            max_len = max([len(_) for _ in clauses])
            self.clauses = torch.zeros([len(clauses), max_len], dtype=torch.int)
            for idx, clause in enumerate(clauses):
                self.clauses[idx][:len(clause)] = torch.tensor(list(clause))
        else:
            self.clauses = torch.zeros(0)


    # @beartype
    def assign(self: 'SATSolver', literal: int | torch.Tensor) -> bool:
        if len(self.clauses) == 0:
            return True

        # save assignment
        if isinstance(literal, torch.Tensor):
            variable = literal.abs().item()
            value = literal.item() > 0
        elif isinstance(literal, int):
            variable = abs(literal)
            value = literal > 0
        else:
            raise NotImplementedError()

        if variable in self.assignment:
            if self.assignment[variable] != (literal > 0):
                return False
            # logger.debug(f'Already assigned: {variable}')
            return True
            # raise ValueError('Already assigned')

        self.assignment[variable] = value
        self._last_assigned_literals = True

        # remove satisfied clauses
        remove_idx = torch.where(self.clauses == literal)[0]
        remain_mask = torch.ones(len(self.clauses), dtype=torch.bool)
        remain_mask[remove_idx] = False
        self.clauses = self.clauses[remain_mask]

        # remove unsatisfied literals
        self.clauses = torch.where(self.clauses == -literal, 0, self.clauses)

        # print('[+] Assigned:', literal, self.assignment)
        # print(self.clauses)
        return True

    # @beartype
    def _bcp_step(self: 'SATSolver') -> tuple[bool, list]:
        if len(self.clauses) == 0:
            return True, []

        inferred_variables = []

        # clauses with single literal -> infer
        literals = self.clauses[torch.where(self.clauses.count_nonzero(dim=1) == 1)[0]].sum(dim=1)
        # print('inferred:', literals)
        for lit in literals:
            inferred_variables.append(lit.item())
            if not self.assign(lit): # conflict
                return False, []
            # print('BCP assign', lit)

        # clauses with 0 literal -> conflict
        if len(torch.where(self.clauses.count_nonzero(dim=1) == 0)[0]):
            return False, []

        return True, inferred_variables

    # @beartype
    def bcp(self: 'SATSolver') -> tuple[bool, list]:
        if len(self.clauses) == 0:
            return True, []

        inferred_variables = []

        stat, inferred = self._bcp_step()
        if not stat:
            return False, []

        inferred_variables += inferred

        # print(self.clauses)
        while self._last_assigned_literals:
            # print('[+] Last assigned:', self._last_assigned_literals)
            self._last_assigned_literals = False
            stat, inferred = self._bcp_step()
            if not stat:
                return False, []
            inferred_variables += inferred

        # print('[+] BCP Infer:', inferred_variables)
        return True, inferred_variables  # no conflict


    def print_stats(self):
        print('[+] Current assignment:', self.assignment)

        print('[+] Remain clauses:')
        print(self.clauses)
        print()


    # @beartype
    def __deepcopy__(self: 'SATSolver', memo: dict) -> SATSolver:
        new_solver = SATSolver([])
        new_solver.clauses = self.clauses.clone()
        new_solver.assignment = copy.deepcopy(self.assignment)
        new_solver._last_assigned_literals = False
        return new_solver


    # @beartype
    def multiple_assign(self: 'SATSolver', literals: list[int]) -> bool:
        if len(self.clauses) == 0:
            return True

        try:
            return self.multiple_assign_cpp(literals)
        except:
            return self.multiple_assign_python(literals)


    # @beartype
    def multiple_assign_python(self: 'SATSolver', literals: list[int]) -> bool:
        if not len(literals):
            return True
        logger.debug(f'[!] Parallel assign using Python: {len(literals)}')

        remain_mask = torch.ones(self.clauses.size(0), dtype=torch.bool, device=self.clauses.device)
        for lit in literals:
            remain_mask &= self.clauses.ne(lit).all(1)
        self.clauses = self.clauses[remain_mask]

        zero_mask = torch.zeros(self.clauses.size(), dtype=torch.bool, device=self.clauses.device)
        for lit in literals:
            zero_mask |= self.clauses.eq(-lit)

        # self.clauses = torch.where(zero_mask, 0, self.clauses)
        self.clauses[zero_mask] = 0
        return True


    # @beartype
    def multiple_assign_cpp(self: 'SATSolver', literals: list[int]) -> bool:
        if not len(literals):
            return True
        import haioc
        logger.debug(f'[!] Parallel assign using C++: {len(literals)}')

        xs = torch.tensor(literals).int()#.cuda()
        # self.clauses = self.clauses.cuda()
        self.clauses = self.clauses[haioc.any_eq_any(self.clauses, xs).logical_not_()]
        self.clauses = haioc.fill_if_eq_any(self.clauses, -xs, 0, inplace=True)
        return True


if __name__ == "__main__":
    # clauses = [
    #     [1, -2, -3 , -4],
    #     [3],
    #     [5, -2, 6],
    #     [1, -3],
    #     [-1, 2, -3],
    #     [-2, -5],
    #     [-1, 6, 5, 7, -2],
    #     # [-1, -3]
    # ]

    # s = SATSolver(clauses)

    # print(s.clauses)
    # logger.setLevel(1)

    # stat, inferred = s.bcp()
    # if not stat:
    #     print('Conflict')
    # else:
    #     print('Infer:', inferred)

    #     s.print_stats()
    # exit()
    seed = random.randint(0, 1000)
    # seed = 549
    print('seed:', seed)
    random.seed(seed)
    lens = list(range(1, 1000+1))
    n_clauses = 20000
    n_vars = 5000
    # clauses = []
    clauses = [random.sample(range(-n_vars, n_vars), random.choice(lens)) for _ in range(n_clauses)]
    clauses = [c for c in clauses if 0 not in c]
    # print(clauses)
    # clauses += [[i, -i] for i in range(1, n_vars+1)]
    lits = [i for i in random.sample(range(-n_vars,n_vars), int(0.5 * n_vars)) if i != 0]
    lits = [l for l in lits if -l not in lits]

    print('Assign literals:', lits[:10])

    ####################
    s = SATSolver(clauses)
    tic = time.time()
    for lit in lits:
        s.assign(lit)
    print('[old] assign:', time.time() - tic)
    s.bcp()
    c1 = s.clauses.clone()

    ####################
    s = SATSolver(clauses)
    tic = time.time()
    s.multiple_assign_python(lits)
    print('[python] assign:', time.time() - tic)
    s.bcp()
    c2 = s.clauses.clone()

    ####################

    s = SATSolver(clauses)
    tic = time.time()
    s.multiple_assign(lits)
    print('[cpp] assign:', time.time() - tic)
    s.bcp()
    c3 = s.clauses.clone()


    ####################
    assert torch.equal(c1, c2)
    assert torch.equal(c2, c3)

    print(c1.shape, 'TRUE')



    # print((s.clauses[..., None] == torch.tensor(lits)[..., None]).nonzero())



""" Reduce operators"""
from .base import *


class BoundReduce(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.axis = attr.get('axes', None)
        self.keepdim = bool(attr['keepdims']) if 'keepdims' in attr else True
        self.use_default_ibp = True

    def _parse_input_and_axis(self, *x):
        if len(x) > 1:
            assert not self.is_input_perturbed(1)
            self.axis = tuple(x[1])
        self.axis = self.make_axis_non_negative(self.axis)
        return x[0]

    def _return_bound_backward(self, lA, uA):
        return [(lA, uA)] + [(None, None)] * (len(self.inputs) - 1), 0, 0


class BoundReduceMax(BoundReduce):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        """Assume that the indexes with the maximum values are not perturbed.
        This generally doesn't hold true, but can still be used for the input shift
        in Softmax of Transformers."""
        self.fixed_max_index = options.get('fixed_reducemax_index', False)

    def _parse_input_and_axis(self, *x):
        x = super()._parse_input_and_axis(*x)
        # for torch.max, `dim` must be an int
        if isinstance(self.axis, tuple):
            assert len(self.axis) == 1
            self.axis = self.axis[0]
        return x

    def forward(self, *x):
        x = self._parse_input_and_axis(*x)
        res = torch.max(x, dim=self.axis, keepdim=self.keepdim)
        self.indices = res.indices
        return res.values

    def bound_backward(self, last_lA, last_uA, *args, **kwargs):
        if self.fixed_max_index:
            def _bound_oneside(last_A):
                if last_A is None:
                    return None
                indices = self.indices.unsqueeze(0)
                if not self.keepdim:
                    assert (self.from_input)
                    last_A = last_A.unsqueeze(self.axis + 1)
                    indices = indices.unsqueeze(self.axis + 1)
                shape = list(last_A.shape)
                shape[self.axis + 1] *= self.input_shape[self.axis]
                A = torch.zeros(shape, device=last_A.device)
                indices = indices.expand(*last_A.shape)
                A.scatter_(dim=self.axis + 1, index=indices, src=last_A)
                return A

            return self._return_bound_backward(_bound_oneside(last_lA),
                                               _bound_oneside(last_uA))
        else:
            raise NotImplementedError(
                '`bound_backward` for BoundReduceMax with perturbed maximum'
                'indexes is not implemented.')


class BoundReduceMin(BoundReduceMax):
    def forward(self, *x):
        x = self._parse_input_and_axis(*x)
        res = torch.min(x, dim=self.axis, keepdim=self.keepdim)
        self.indices = res.indices
        return res.values


class BoundReduceMean(BoundReduce):
    def forward(self, *x):
        x = self._parse_input_and_axis(*x)
        return torch.mean(x, dim=self.axis, keepdim=self.keepdim)

    def bound_backward(self, last_lA, last_uA, *args, **kwargs):
        def _bound_oneside(last_A):
            if last_A is None:
                return None
            if not self.keepdim:
                assert (self.from_input)
                for axis in self.axis:
                    if axis > 0:
                        last_A = last_A.unsqueeze(axis + 1)
            for axis in self.axis:
                shape = list(last_A.shape)
                size_axis = self.input_shape[axis]
                shape[axis + 1] *= size_axis
                last_A = last_A.expand(*shape) / size_axis
            return last_A

        return self._return_bound_backward(_bound_oneside(last_lA),
                                           _bound_oneside(last_uA))

    def bound_forward(self, dim_in, x, *args):
        assert self.keepdim
        assert (len(self.axis) == 1)
        axis = self.make_axis_non_negative(self.axis[0])
        assert (axis > 0)
        size = self.input_shape[axis]
        lw = x.lw.sum(dim=axis + 1, keepdim=True) / size
        lb = x.lb.sum(dim=axis, keepdim=True) / size
        uw = x.uw.sum(dim=axis + 1, keepdim=True) / size
        ub = x.ub.sum(dim=axis, keepdim=True) / size
        return LinearBound(lw, lb, uw, ub)


class BoundReduceSum(BoundReduce):
    def forward(self, *x):
        x = self._parse_input_and_axis(*x)
        if self.axis is not None:
            return torch.sum(x, dim=self.axis, keepdim=self.keepdim)
        else:
            return torch.sum(x)

    def bound_backward(self, last_lA, last_uA, x, *args, **kwargs):
        def _bound_oneside(last_A):
            if last_A is None:
                return None
            if not self.keepdim:
                assert (self.from_input)
                for axis in self.axis:
                    if axis > 0:
                        last_A = last_A.unsqueeze(axis + 1)
            for axis in self.axis:
                shape = list(last_A.shape)
                shape[axis + 1] *= self.input_shape[axis]
                last_A = last_A.expand(*shape)
            return last_A

        return self._return_bound_backward(_bound_oneside(last_lA),
                                           _bound_oneside(last_uA))

    def bound_forward(self, dim_in, x, *args):
        assert len(self.axis) == 1
        axis = self.make_axis_non_negative(self.axis[0])
        assert axis > 0
        lw = x.lw.sum(dim=axis + 1, keepdim=self.keepdim)
        lb = x.lb.sum(dim=axis, keepdim=self.keepdim)
        uw = x.uw.sum(dim=axis + 1, keepdim=self.keepdim)
        ub = x.ub.sum(dim=axis, keepdim=self.keepdim)
        return LinearBound(lw, lb, uw, ub)



""" Softmax """
from .base import *

class BoundSoftmaxImpl(nn.Module):
    def __init__(self, axis):
        super().__init__()
        self.axis = axis
        assert self.axis == int(self.axis)

    def forward(self, x):
        max_x = torch.max(x, dim=self.axis).values
        x = torch.exp(x - max_x.unsqueeze(self.axis))
        s = torch.sum(x, dim=self.axis, keepdim=True)
        return x / s

# The `option != 'complex'` case is not used in the auto_LiRPA main paper.
class BoundSoftmax(Bound):
    def __init__(self, attr=None, inputs=None, output_index=0, options=None):
        super().__init__(attr, inputs, output_index, options)
        self.axis = attr['axis']
        self.option = options.get('softmax', 'complex')
        if self.option == 'complex':
            self.complex = True
        else:
            self.max_input = 30

    def forward(self, x):
        assert self.axis == int(self.axis)
        if self.option == 'complex':
            self.input = (x,)
            self.model = BoundSoftmaxImpl(self.axis)
            self.model.device = self.device
            return self.model(x)
        else:
            return F.softmax(x, dim=self.axis)

    def interval_propagate(self, *v):
        assert self.option != 'complex'
        assert self.perturbed
        h_L, h_U = v[0]
        shift = h_U.max(dim=self.axis, keepdim=True).values
        exp_L, exp_U = torch.exp(h_L - shift), torch.exp(h_U - shift)
        lower = exp_L / (torch.sum(exp_U, dim=self.axis, keepdim=True) - exp_U + exp_L + epsilon)
        upper = exp_U / (torch.sum(exp_L, dim=self.axis, keepdim=True) - exp_L + exp_U + epsilon)
        return lower, upper



import torch
from torch import nn

from ..utils import PRINT_DEBUG

class Gather(nn.Module):
    def __init__(self, dim=0):
        self.dim = dim
        self.selection = [slice(None) for _ in range(dim)]
        super().__init__()

    def forward(self, data: torch.Tensor, indices: torch.Tensor):
        # selection = self.selection + [indices.to(torch.int64)]
        # return data.__getitem__(selection)
        if PRINT_DEBUG:
            print('GATHER:', data.shape, data.is_floating_point())
            # if data.ndim == 1:
            #     data = data[None]

            print('\t- self.dim:', self.dim)
            print('\t- indices:', indices)

        if data.is_floating_point():
            dim = max(self.dim, 1)
        else:
            dim = self.dim


        if indices.numel() == 1 and indices == -1:
            indices = torch.tensor(data.shape[dim] - 1, device=data.device)

        # if indices.ndim == 0:
        #     final = torch.index_select(data, dim=dim, index=indices)
        #     # indices = indices.view(-1)#.repeat(data.shape[0])
        #     print('\t- final 1:', final.shape)
        #     # final = final.squeeze(dim)
        #     # self.dim += 1
        # else:
        #     final = torch.index_select(data, dim=dim, index=indices)
        #     raise

        final = torch.index_select(data, dim=dim, index=indices).squeeze(dim)

        if PRINT_DEBUG:
            print('\t- dim:', dim)
            print('\t- final:', final)
            print('\t- final:', final.shape)
            assert final.numel() > 0
        # exit()
        return final



from __future__ import annotations
from collections import defaultdict
from beartype import beartype
import numpy as np
import random
import typing
import torch
import time
import copy
import os

if typing.TYPE_CHECKING:
    from .. import verifier

from ..heuristic.restart_heuristics import get_restart_strategy, HIDDEN_SPLIT_RESTART_STRATEGIES
from ..heuristic.util import compute_masks, _history_to_clause
from ..heuristic.decision_heuristics import DecisionHeuristic
from ..heuristic.tightener import Tightener

from ..attacker.pgd_attack.general import general_attack
from ..attacker.mip_attack import MIPAttacker
from ..attacker.attacker import Attacker

from ..abstractor.abstractor import NetworkAbstractor

from ..util.misc.result import AbstractResults, ReturnStatus
from ..util.misc.check import check_solution
from ..util.misc.logger import logger


from ..setting import Settings


# @beartype
def _prune_domains(domain_params: AbstractResults, remaining_indices: torch.Tensor) -> AbstractResults:
    return AbstractResults(**{
        'objective_ids': domain_params.objective_ids[remaining_indices],
        'output_lbs': domain_params.output_lbs[remaining_indices] if domain_params.output_lbs is not None else None,
        'masks': {k: v[remaining_indices] for k, v in domain_params.masks.items()} if domain_params.masks is not None else None,
        'lAs': {k: v[remaining_indices] for k, v in domain_params.lAs.items()},
        'histories': [domain_params.histories[_] for _ in remaining_indices],
        'lower_bounds': {k: v[remaining_indices] for k, v in domain_params.lower_bounds.items()},
        'upper_bounds': {k: v[remaining_indices] for k, v in domain_params.upper_bounds.items()},
        'input_lowers': domain_params.input_lowers[remaining_indices],
        'input_uppers': domain_params.input_uppers[remaining_indices],
        'betas': [domain_params.betas[_] for _ in remaining_indices],
        'cs': domain_params.cs[remaining_indices],
        'rhs': domain_params.rhs[remaining_indices],
        'sat_solvers': [domain_params.sat_solvers[_] for _ in remaining_indices] if domain_params.sat_solvers is not None else None,
        'slopes': defaultdict(dict, {k: {kk: vv[:, :, remaining_indices] for kk, vv in v.items()} for k, v in domain_params.slopes.items()}),
    })


# @beartype
def _mip_attack(self, reference_bounds: dict | None) -> tuple[bool, torch.Tensor | None]:
    if not Settings.use_attack:
        return False, None

    if not Settings.use_mip_attack:
        return False, None

    return self.mip_attacker.run(reference_bounds)


# @beartype
def _preprocess(self, objectives: typing.Any, force_split: str | None = None) -> tuple:
    # determine search algorithm
    self.refined_betas = None

    diff = objectives.upper_bounds - objectives.lower_bounds
    eps = diff.max().item()
    perturbed = (diff > 0).int().sum()
    logger.info(f'[!] eps={eps:.06f}, perturbed={perturbed}')

    if Settings.test:
        self.input_split = False
    elif force_split is not None:
        assert force_split in ['input', 'hidden']
        self.input_split = force_split == 'input'
    elif eps > Settings.safety_property_threshold: # safety properties
        self.input_split = True
    elif np.prod(self.input_shape) <= 200: # small inputs
        self.input_split = True
    elif np.prod(self.input_shape) >= 100000: # large inputs, e.g., VGG16
        self.input_split = True

    if self.input_split:
        return objectives, None

    if (not isinstance(objectives.cs, torch.Tensor)) or (not isinstance(objectives.rhs, torch.Tensor)):
        return objectives, None

    if not torch.allclose(objectives.lower_bounds.mean(dim=0), objectives.lower_bounds[0], 1e-5, 1e-5):
        return objectives, None

    try:
        # self._init_abstractor('crown-optimized', objectives)
        self._init_abstractor('backward' if np.prod(self.input_shape) < 100000 else 'forward', objectives)
    except:
        raise
        print('Failed to initialize abstractor')
        return objectives, None

    # prune objectives
    tmp_objective = copy.deepcopy(objectives)
    tmp_objective.lower_bounds = tmp_objective.lower_bounds[0:1] # raise errors if using beta, use full objectives instead
    tmp_objective.upper_bounds = tmp_objective.upper_bounds[0:1] # raise errors if using beta, use full objectives instead

    # forward
    try:
        ret = self.abstractor.initialize(tmp_objective)
    except:
        print('Failed to initialize objectives')
        return objectives, None

    # pruning
    remaining_index = torch.where((ret.output_lbs.detach().cpu() <= tmp_objective.rhs.detach().cpu()).all(1))[0]
    objectives.lower_bounds = objectives.lower_bounds[remaining_index]
    objectives.upper_bounds = objectives.upper_bounds[remaining_index]
    objectives.cs = objectives.cs[remaining_index]
    objectives.rhs = objectives.rhs[remaining_index]
    objectives.lower_bounds_f64 = objectives.lower_bounds_f64[remaining_index]
    objectives.upper_bounds_f64 = objectives.upper_bounds_f64[remaining_index]
    objectives.cs_f64 = objectives.cs_f64[remaining_index]
    objectives.rhs_f64 = objectives.rhs_f64[remaining_index]

    if None in self.abstractor.split_points:
        # FIXME: disable restart + stabilize for now
        Settings.use_restart = False
        Settings.use_mip_tightening = False
        logger.info(f'Remain {len(objectives)} objectives')
        return objectives, None

    # refine
    refined_intermediate_bounds = None
    if len(objectives) and (Settings.use_mip_tightening) and self.abstractor.method == 'backward':
        use_refined = not Settings.use_restart
        if any([isinstance(_, (torch.nn.Conv1d, torch.nn.Conv2d, torch.nn.Conv3d,
                               torch.nn.ConvTranspose1d, torch.nn.ConvTranspose2d, torch.nn.ConvTranspose3d))
                            for _ in self.net.modules()][1:]):
            # skip refine for Conv layers for now
            use_refined = False

        if use_refined:
            logger.info(f'Refining hidden bounds for {len(objectives)} remaining objectives')
            tmp_objective = copy.deepcopy(objectives)
            tmp_objective.lower_bounds = tmp_objective.lower_bounds[0:1].to(self.device)
            tmp_objective.upper_bounds = tmp_objective.upper_bounds[0:1].to(self.device)

            # build solver
            tic = time.time()
            c_to_use = tmp_objective.cs.transpose(0, 1).to(self.device) if tmp_objective.cs.shape[1] == 1 else None
            self.abstractor.build_lp_solver(
                model_type='mip',
                input_lower=tmp_objective.lower_bounds.view(self.input_shape),
                input_upper=tmp_objective.upper_bounds.view(self.input_shape),
                c=c_to_use,
                refine=use_refined,
                timeout=None,
                timeout_per_neuron=Settings.mip_tightening_timeout_per_neuron,
            )
            logger.debug(f'MIP: {time.time() - tic:.04f}')

            # forward with refinement
            refined_intermediate_bounds = self.abstractor.net.get_refined_interm_bounds()
            ret = self.abstractor.initialize(tmp_objective, reference_bounds=refined_intermediate_bounds)

            # pruning
            remaining_index = torch.where((ret.output_lbs.detach().cpu() <= tmp_objective.rhs.detach().cpu()).all(1))[0]
            objectives.lower_bounds = objectives.lower_bounds[remaining_index]
            objectives.upper_bounds = objectives.upper_bounds[remaining_index]
            objectives.cs = objectives.cs[remaining_index]
            objectives.rhs = objectives.rhs[remaining_index]
            objectives.lower_bounds_f64 = objectives.lower_bounds_f64[remaining_index]
            objectives.upper_bounds_f64 = objectives.upper_bounds_f64[remaining_index]
            objectives.cs_f64 = objectives.cs_f64[remaining_index]
            objectives.rhs_f64 = objectives.rhs_f64[remaining_index]

            # TODO: fixme (update found betas from MIP)
            # self.refined_betas = self.abstractor.net.get_betas()

        # torch.save(refined_intermediate_bounds, 'refined.pt')

    # mip tightener
    if len(objectives) and (not self.input_split):
        # mip attacker
        if Settings.use_mip_attack:
            self.mip_attacker = MIPAttacker(
                abstractor=self.abstractor,
                objectives=objectives, # full objectives
            )

        if Settings.use_mip_tightening:
            self.tightener = Tightener(
                abstractor=self.abstractor,
                objectives=objectives,
            )

    logger.info(f'Remain {len(objectives)} objectives')
    # refined_intermediate_bounds = torch.load('refined.pt')
    return objectives, refined_intermediate_bounds

# @beartype
def _check_timeout(self, timeout: float) -> bool:
    return time.time() - self.start_time > timeout


# @beartype
def _init_abstractor(self, method: str, objective: typing.Any) -> None:
    if hasattr(self, 'abstractor'):
        # del self.abstractor.net
        del self.abstractor

    self.abstractor = NetworkAbstractor(
        pytorch_model=self.net,
        input_shape=self.input_shape,
        method=method,
        input_split=self.input_split,
        device=self.device,
    )

    self.abstractor.setup(objective)
    self.abstractor.net.get_split_nodes(input_split=False)


# @beartype
def _setup_restart(self, nth_restart: int, objective: typing.Any) -> None | dict:
    self.num_restart = nth_restart + 1
    params = get_restart_strategy(nth_restart, input_split=self.input_split)
    if params is None:
        raise NotImplementedError()

    if np.prod(self.input_shape) >= 100000: # large inputs, e.g., VGG16
        Settings.forward_dynamic = True
        Settings.forward_max_dim = 100
        Settings.backward_batch_size = 16

    logger.info(f'Params of {nth_restart+1}-th run: {params}')
    abstract_method = params['abstract_method']

    # decision heuristic
    assert params['input_split'] == self.input_split
    self.decision = DecisionHeuristic(
        input_split=params['input_split'],
        decision_topk=params['decision_topk'],
        decision_method=params['decision_method'],
    )

    refined_intermediate_bounds = None
    if (not self.input_split) and Settings.use_restart and self.num_restart == len(HIDDEN_SPLIT_RESTART_STRATEGIES) and Settings.use_mip_tightening:
        if 'forward' in abstract_method:
            pass
        elif not torch.allclose(objective.lower_bounds.mean(dim=0), objective.lower_bounds[0], 1e-5, 1e-5):
            pass
        elif any([isinstance(_, (torch.nn.Conv2d, torch.nn.Conv3d, torch.nn.ConvTranspose2d, torch.nn.ConvTranspose3d)) for _ in self.net.modules()][1:]):
            # FIXME: skip refine for Conv layers
            pass
        elif None in self.abstractor.split_points:
            # skip refine for general activation layers
            pass
        else:
            self._init_abstractor('backward', objective)

            tmp_objective = copy.deepcopy(objective)
            tmp_objective.lower_bounds = tmp_objective.lower_bounds[0:1].to(self.device)
            tmp_objective.upper_bounds = tmp_objective.upper_bounds[0:1].to(self.device)
            # tmp_objective.rhs = tmp_objective.rhs.to(self.device) # TODO: check
            c_to_use = tmp_objective.cs.transpose(0, 1).to(self.device) if tmp_objective.cs.shape[1] == 1 else None

            ret = self.abstractor.initialize(tmp_objective)
            self.abstractor.build_lp_solver(
                model_type='mip',
                input_lower=tmp_objective.lower_bounds.view(self.input_shape),
                input_upper=tmp_objective.upper_bounds.view(self.input_shape),
                c=c_to_use,
                refine=True,
                timeout=None,
                timeout_per_neuron=Settings.mip_tightening_timeout_per_neuron,
            )
            refined_intermediate_bounds = self.abstractor.net.get_refined_interm_bounds()

    # abstractor
    if (not hasattr(self, 'abstractor')) or (abstract_method != self.abstractor.method):
        self._init_abstractor(abstract_method, objective)

    return refined_intermediate_bounds


# @beartype
def _pre_attack(self, dnf_objectives: verifier.objective.DnfObjectives,
                timeout: float = 2.0) -> tuple[bool, torch.Tensor | None]:
    if Settings.use_attack:
        return Attacker(self.net, dnf_objectives, self.input_shape, device=self.device).run(timeout=timeout)
    return False, None

# @beartype
def _random_idx(total_samples: int, num_samples: int, device: str = 'cpu') -> torch.Tensor:
    if num_samples >= total_samples:
        return torch.Tensor(range(total_samples)).to(device)
    return torch.Tensor(random.sample(range(total_samples), num_samples)).to(device)


# @beartype
def _attack(self, domain_params: AbstractResults, timeout: float,
            n_sample: int = 50, n_interval: int = 1) -> torch.Tensor | None:
    if not Settings.use_attack:
        return None

    if self.iteration % n_interval != 0:
        return None

    # random samples
    indices = _random_idx(len(domain_params.cs), n_sample, device=self.device).long()

    input_lowers = domain_params.input_lowers[indices][None]
    input_uppers = domain_params.input_uppers[indices][None]
    # adv_example = (input_lowers + input_uppers) / 2
    adv_example = (input_uppers - input_lowers) * torch.rand(input_lowers.shape, device=self.device) + input_lowers
    if os.environ.get('NEURALSAT_ASSERT'):
        assert torch.all(adv_example <= input_uppers)
        assert torch.all(adv_example >= input_lowers)

    cs = domain_params.cs[indices].view(1, -1, domain_params.cs[indices].shape[-1])
    rhs = domain_params.rhs[indices].view(1, -1)
    cond = [[domain_params.cs.shape[1] for i in range(len(indices))]]
    serialized_conditions = (cs, rhs, cond)

    attack_images = general_attack(
        model=self.net,
        X=adv_example,
        data_min=input_lowers,
        data_max=input_uppers,
        serialized_conditions=serialized_conditions,
        attack_iters=20,
        num_restarts=5,
        only_replicate_restarts=True,
        use_gama=False,
        timeout=timeout,
    )
    if (attack_images is None) and (self.iteration % (5 * n_interval) == 0) and 0:
        attack_images = general_attack(
            model=self.net,
            X=adv_example,
            data_min=input_lowers,
            data_max=input_uppers,
            serialized_conditions=serialized_conditions,
            attack_iters=30,
            num_restarts=10,
            only_replicate_restarts=True,
            use_gama=True,
            timeout=timeout,
        )

    if attack_images is not None:
        for i in range(attack_images.shape[1]): # restarts
            for j in range(attack_images.shape[2]): # props
                adv = attack_images[:, i, j]
                if check_solution(self.net, adv, domain_params.cs[indices][j], domain_params.rhs[indices][j], input_lowers[:, j], input_uppers[:, j]):
                    return adv
        logger.debug("[!] Invalid counter-example")

    return None


# @beartype
def _get_learned_conflict_clauses(self) -> dict:
    if hasattr(self.domains_list, 'all_conflict_clauses'):
        return self.domains_list.all_conflict_clauses
    return {}


# @beartype
def _check_invoke_tightening(self, patience_limit: int = 10):
    if not Settings.use_mip_tightening:
        return False

    if Settings.test:
        return True

    if self.input_split:
        return False

    if self.tightening_patience < patience_limit:
        return False

    if len(self.domains_list) <= self.batch:
        return False

    if Settings.use_restart and self.num_restart < len(HIDDEN_SPLIT_RESTART_STRATEGIES):
        return False

    # reset counter
    self.tightening_patience = 0
    return True


# @beartype
def _update_tightening_patience(self, minimum_lowers: float, old_domains_length: int) -> None:
    current_domains_length = len(self.domains_list)
    if (minimum_lowers > self.last_minimum_lowers) or (current_domains_length <= self.batch):
        self.tightening_patience -= 1
        # self.tightening_patience = 0
    elif (current_domains_length <= old_domains_length):
        self.tightening_patience -= 1
    elif minimum_lowers == self.last_minimum_lowers:
        self.tightening_patience += 1
    else:
        self.tightening_patience += 3

    self.tightening_patience = max(0, self.tightening_patience)
    self.last_minimum_lowers = minimum_lowers


# @beartype
def _check_full_assignment(self, domain_params: AbstractResults) -> tuple[torch.Tensor | None, torch.Tensor | None]:
    if self.input_split:
        return None, None

    if domain_params.lower_bounds is None:
        return None, None

    # check all activation layers are ReLU
    if None in self.abstractor.split_points:
        # general activation
        return None, None

    new_masks = compute_masks(
        lower_bounds=domain_params.lower_bounds,
        upper_bounds=domain_params.upper_bounds,
        device='cpu', non_blocking=False, # FIXME: setting True makes it return wrong values
    )

    n_unstables = torch.stack([v.sum(dim=1) for k, v in new_masks.items()]).sum(dim=0)
    pruning_indices = torch.where(n_unstables == 0)[0]

    if not len(pruning_indices):
        return None, None

    for idx_ in pruning_indices:
        self.abstractor.build_lp_solver(
            model_type='lp',
            input_lower=domain_params.input_lowers[idx_][None],
            input_upper=domain_params.input_uppers[idx_][None],
            c=domain_params.cs[idx_][None],
            refine=False,
        )

        feasible, adv = self.abstractor.solve_full_assignment(
            input_lower=domain_params.input_lowers[idx_],
            input_upper=domain_params.input_uppers[idx_],
            lower_bounds={k: v[idx_] for k, v in domain_params.lower_bounds.items()},
            upper_bounds={k: v[idx_] for k, v in domain_params.upper_bounds.items()},
            c=domain_params.cs[idx_],
            rhs=domain_params.rhs[idx_]
        )

        if feasible:
            return adv, None

    # save pruned domains
    [self.domains_list.all_conflict_clauses[int(domain_params.objective_ids[i])].append(domain_params.histories[i]) for i in pruning_indices]

    # unverified indices
    remaining_indices = torch.where(n_unstables > 0)[0]

    return None, remaining_indices


# @beartype
def compute_stability(self, dnf_objectives: verifier.objective.DnfObjectives) -> tuple[int, int, list, list]:
    if not (hasattr(self, 'abstractor')):
        self._init_abstractor('backward' if np.prod(self.input_shape) < 100000 else 'forward', dnf_objectives)

    return self.abstractor.compute_stability(dnf_objectives)


# @beartype
def _save_stats(self) -> None:
    for k, v in self._get_learned_conflict_clauses().items():
        if k not in self.all_conflict_clauses:
            self.all_conflict_clauses[k] = []
        self.all_conflict_clauses[k].extend(v)
    if hasattr(self.domains_list, 'visited'):
        self.visited += self.domains_list.visited


# @beartype
def get_stats(self) -> tuple[int, int]:
    depths = {}
    for k, v in self.all_conflict_clauses.items():
        depths[k] = max(map(lambda x: sum(len(_[0]) for _ in x.values()), v)) if len(v) else 0
    depth = max(list(depths.values())) if len(depths) else 0
    return depth, self.visited


# @beartype
def _check_adv_f64(self, adv: torch.Tensor, objective: typing.Any) -> bool:
    lower_bounds_f64 = objective.lower_bounds_f64.view(-1, *self.input_shape[1:]).to(self.device)
    upper_bounds_f64 = objective.upper_bounds_f64.view(-1, *self.input_shape[1:]).to(self.device)
    cs_f64 = objective.cs_f64.to(self.device)
    rhs_f64 = objective.rhs_f64.to(self.device)
    for i in range(len(lower_bounds_f64)):
        if check_solution(self.net, adv, cs_f64[i], rhs_f64[i], lower_bounds_f64[i:i+1], upper_bounds_f64[i:i+1]):
            return True
    return False


# @beartype
def get_unsat_core(self) -> None | dict:
    if self.status != ReturnStatus.UNSAT:
        return None

    unsat_cores = {k: [] for k in self.all_conflict_clauses}
    for k, v in self.all_conflict_clauses.items():
        [unsat_cores[k].append(_history_to_clause(c, self.domains_list.var_mapping)) for c in v]

    return unsat_cores



import argparse
import torch
import time
import os

from util.misc.logger import logger, LOGGER_LEVEL
from util.spec.read_vnnlib import read_vnnlib
from util.network.read_onnx import parse_onnx
from util.misc.export import get_adv_string
from util.misc.timer import Timers

from verifier.objective import Objective, DnfObjectives
from verifier.verifier import Verifier 

from setting import Settings


def print_w_b(model):
    for layer in model.modules():
        if hasattr(layer, 'weight'):
            print(layer)
            print('\t[+] w:', layer.weight.data.detach().flatten())
            print('\t[+] b:', layer.bias.data.detach().flatten())
            print()
            
if __name__ == '__main__':
    START_TIME = time.time()

    # argument
    parser = argparse.ArgumentParser()
    parser.add_argument('--net', type=str, required=True,
                        help="load pretrained ONNX model from this specified path.")
    parser.add_argument('--spec', type=str, required=True,
                        help="path to VNNLIB specification file.")
    parser.add_argument('--batch', type=int, default=1000,
                        help="maximum number of branches to verify in each iteration")
    parser.add_argument('--timeout', type=float, default=3600,
                        help="timeout in seconds")
    parser.add_argument('--device', type=str, default='cuda', choices=['cpu', 'cuda'],
                        help="choose device to use for verifying.")
    parser.add_argument('--verbosity', type=int, choices=[0, 1, 2], default=2, 
                        help='the logger level (0: NOTSET, 1: INFO, 2: DEBUG).')
    parser.add_argument('--result_file', type=str, required=False,
                        help="file to save execution results.")
    parser.add_argument('--export_cex', action='store_true',
                        help="export counter-example to result file.")
    parser.add_argument('--disable_restart', action='store_false',
                        help="disable RESTART heuristic.")
    parser.add_argument('--disable_stabilize', action='store_false',
                        help="disable STABILIZE heuristic.")
    parser.add_argument('--force_split', type=str, choices=['input', 'hidden'],
                        help="select SPLITTING strategy.")
    parser.add_argument('--test', action='store_true',
                        help="test on small example with special settings.")
    args = parser.parse_args()   
    
    
    # setup timers
    if Settings.use_timer:
        Timers.reset()
        Timers.tic('Main')
        
    # set device
    if not torch.cuda.is_available():
        args.device = 'cpu'
        
    if args.test:
        Settings.setup_test()
    else:
        Settings.setup(args)
        
    # set logger level
    logger.setLevel(LOGGER_LEVEL[args.verbosity])
    
    # network
    Timers.tic('Load network') if Settings.use_timer else None
    model, input_shape, output_shape, is_nhwc = parse_onnx(args.net)
    model.to(args.device)
    Timers.toc('Load network') if Settings.use_timer else None
    
    if args.verbosity:
        print(model)
        if Settings.test:
            print_w_b(model)
    
    # specification
    Timers.tic('Load specification') if Settings.use_timer else None
    vnnlibs = read_vnnlib(args.spec)
    logger.info(f'[!] Input shape: {input_shape} (is_nhwc={is_nhwc})')
    logger.info(f'[!] Output shape: {output_shape}')
    Timers.toc('Load specification') if Settings.use_timer else None
    
    # verifier
    verifier = Verifier(
        net=model, 
        input_shape=input_shape, 
        batch=args.batch,
        device=args.device,
    )
    
    # objective
    objectives = []
    for spec in vnnlibs:
        bounds = spec[0]
        for prop_i in spec[1]:
            objectives.append(Objective((bounds, prop_i)))
            
    objectives = DnfObjectives(
        objectives=objectives, 
        input_shape=input_shape, 
        is_nhwc=is_nhwc,
    )
    
    print(Settings)
    
    # verify
    Timers.tic('Verify') if Settings.use_timer else None
    timeout = args.timeout - (time.time() - START_TIME)
    status = verifier.verify(objectives, timeout=timeout, force_split=args.force_split)
    runtime = time.time() - START_TIME
    Timers.toc('Verify') if Settings.use_timer else None
    
    # output
    logger.info(f'[!] Iterations: {verifier.iteration}')
    if verifier.adv is not None:
        logger.info(f'adv (first 5): {verifier.adv.flatten()[:5].detach().cpu()}')
        logger.debug(f'output: {verifier.net(verifier.adv).flatten().detach().cpu()}')
        
    # export
    if args.result_file:
        os.remove(args.result_file) if os.path.exists(args.result_file) else None
        with open(args.result_file, 'w') as fp:
            print(f'{status},{runtime:.06f}', file=fp)
            if (verifier.adv is not None) and args.export_cex:
                print(get_adv_string(inputs=verifier.adv, net_path=args.net, is_nhwc=is_nhwc), file=fp)

    logger.info(f'[!] Result: {status}')
    logger.info(f'[!] Runtime: {runtime:.04f}')
    # logger.debug(f'[!] UNSAT core: {verifier.get_unsat_core()}')
    
    if Settings.use_timer:
        Timers.toc('Main')
        Timers.print_stats()
        
    print(f'{status},{runtime:.04f}')



import numpy as np
import torch


def scale_saliency_map_np(image_3d, percentile=99, dim=0):
    r"""Returns a 3D tensor as a grayscale 2D tensor.
    This method sums a 3D tensor across the absolute value of axis=2, and then
    clips values at a given percentile.
    """
    image_2d = np.sum(np.abs(image_3d), axis=dim)

    vmax = np.percentile(image_2d, percentile)
    vmin = np.min(image_2d)

    return np.clip((image_2d - vmin) / (vmax - vmin), 0, 1)

def diverge_saliency_map_np(image_3d, percentile=99, dim=0):
    r"""Returns a 3D tensor as a 2D tensor with positive and negative values.
    """
    image_2d = np.sum(image_3d, axis=dim)

    span = abs(np.percentile(image_2d, percentile))
    vmin = -span
    vmax = span

    return np.clip((image_2d - vmin) / (vmax - vmin), -1, 1)


class IntegratedGradient:

    "Attribution method"

    def __init__(self, model, device):
        self.model = model.to(device)
        self.device = device


    def attribute(self, x, steps=20, baseline=None):
        x = x.to(self.device)

        # baselines
        if baseline is None:
            # baseline = torch.randn(x.shape, requires_grad=False).to(x.device)
            baseline = torch.ones_like(x, requires_grad=False, device=self.device) * x.mean()
        self._check(x, baseline)

        # predict label
        output = self.model(x)
        pred_label = output.max(1)[1]

        # inputs
        x.requires_grad_(True)
        X, delta_X = self._get_X_and_delta(x, baseline, steps)

        # integrated gradients
        grad = torch.autograd.grad(self.model(X)[:, pred_label].sum(), X)[0]
        # grad = grad.cpu().numpy()
        grad = delta_X * (grad[:-1] + grad[1:]) / 2.
        ig_grad = grad.sum(dim=0, keepdims=True)

        return ig_grad, pred_label[0]


    def _get_X_and_delta(self, x, baseline, steps):
        alphas = torch.linspace(0, 1, steps + 1).view(-1, 1, 1, 1).to(x.device)
        delta = (x - baseline)
        x = baseline + alphas * delta
        return x, (delta / steps).detach()


    def _check(self, x, baseline):
        if x.shape != baseline.shape:
            raise ValueError(f'input shape should equal to baseline shape. '
                             f'Got input shape: {x.shape}, '
                             f'baseline shape: {baseline.shape}')



from beartype import beartype

from ..setting import Settings

INPUT_SPLIT_RESTART_STRATEGIES = [
    {'input_split': True, 'abstract_method': 'forward+backward', 'decision_method': 'smart', 'decision_topk': 1},
    {'input_split': True, 'abstract_method': 'forward+backward', 'decision_method': 'naive', 'decision_topk': 1},

    {'input_split': True, 'abstract_method': 'backward',         'decision_method': 'smart', 'decision_topk': 1},
    {'input_split': True, 'abstract_method': 'backward',         'decision_method': 'naive', 'decision_topk': 1},

    {'input_split': True, 'abstract_method': 'crown-optimized',  'decision_method': 'smart', 'decision_topk': 1},
    {'input_split': True, 'abstract_method': 'crown-optimized',  'decision_method': 'naive', 'decision_topk': 1},
]

HIDDEN_SPLIT_RESTART_STRATEGIES = [
    # {'input_split': False, 'abstract_method': 'crown-optimized', 'decision_method': 'brute-force', 'decision_topk': -1},
    {'input_split': False, 'abstract_method': 'crown-optimized', 'decision_method': 'smart', 'decision_topk': 10},
    # {'input_split': False, 'abstract_method': 'crown-optimized', 'decision_method': 'smart', 'decision_topk': 20},
    {'input_split': False, 'abstract_method': 'crown-optimized', 'decision_method': 'smart', 'decision_topk': 30},
]


# @beartype
def get_restart_strategy(nth_restart: int, input_split: bool = False) -> dict:
    if input_split:
        if not Settings.use_restart:
            strategy = INPUT_SPLIT_RESTART_STRATEGIES[3] # default: backward + naive
        elif nth_restart >= len(INPUT_SPLIT_RESTART_STRATEGIES):
            strategy = INPUT_SPLIT_RESTART_STRATEGIES[-1]
        else:
            strategy = INPUT_SPLIT_RESTART_STRATEGIES[nth_restart]
    else:
        if nth_restart >= len(HIDDEN_SPLIT_RESTART_STRATEGIES):
            strategy = HIDDEN_SPLIT_RESTART_STRATEGIES[-1]
        else:
            strategy = HIDDEN_SPLIT_RESTART_STRATEGIES[nth_restart]
    return strategy



import gurobipy as grb
import multiprocessing
import numpy as np
import random
import torch
import copy
import time
import sys
import os

from ..util.misc.logger import logger
from ..setting import Settings

MULTIPROCESS_MODEL = None
DEBUG = True

REMOVE_UNUSED = True

def _bound_improvement(orig, refined, bound_type):
    assert len(orig) == len(refined)
    if bound_type == 'lower':
        assert all([(i <= ii).all() for (i, ii) in zip(orig, refined)])
        return [(ii - i).sum() for (i, ii) in zip(orig, refined)]
    assert all([(i >= ii).all() for (i, ii) in zip(orig, refined)])
    return [(i - ii).sum() for (i, ii) in zip(orig, refined)]


def handle_gurobi_error(message):
    print(f'Gurobi error: {message}')
    raise NotImplementedError()


def _get_prefix_constr_name(name):
    if name.startswith('lay'):
        return ''.join(name.split('_')[:-2])
    return ''.join(name.split('_')[:-3])


def _get_prefix_var_name(name):
    return ''.join(name.split('_')[:-1])


def mip_solver_worker(candidate):
    """ Multiprocess worker for solving MIP models in build_the_model_mip_refine """

    def remove_unused_vars_and_constrs(grb_model, var_name, pre_activation_names, activation_names, final_name):
        # print('removing some variables and constraints', var_name, pre_activation_names, activation_names)

        current_layer_name = ''.join(var_name.split('_')[:-1])[3:] # remove "lay" and "_{nid}"
        assert current_layer_name in pre_activation_names
        current_layer_id = pre_activation_names[current_layer_name]
        # print('current_layer_name:', current_layer_name)
        # print('current_layer_id:', current_layer_id)

        remove_pre_activation_patterns = [f'lay{k}' for k, v in pre_activation_names.items() if v >= current_layer_id]
        remove_pre_activation_patterns += [f'lay{final_name}']
        remove_activation_patterns = [f'ReLU{v}' for k, v in activation_names.items() if k >= current_layer_id]
        remove_activation_patterns += [f'aReLU{v}' for k, v in activation_names.items() if k >= current_layer_id]
        # print('remove_pre_activation_patterns:', remove_pre_activation_patterns)
        # print('remove_activation_patterns:', remove_activation_patterns)
        all_remove_patterns = remove_pre_activation_patterns + remove_activation_patterns
        remove_vars = []
        remove_constrs = []

        # remove constraints
        for c_ in grb_model.getConstrs():
            if c_.ConstrName == f'{var_name}_eq':
                # print('skip', c_.ConstrName)
                continue
            if _get_prefix_constr_name(c_.ConstrName) in all_remove_patterns:
                remove_constrs.append(c_)
                # remove_constrs.append(c_.ConstrName)
            # print(c_.ConstrName, _get_prefix_constr_name(c_.ConstrName), )

        # remove variables
        for v_ in grb_model.getVars():
            if v_.VarName == var_name:
                # print('skip', var_name)
                continue
            if _get_prefix_var_name(v_.VarName) in all_remove_patterns:
                remove_vars.append(v_)
                # remove_vars.append(v_.VarName)
            # print(v_.VarName)

        grb_model.remove(remove_constrs)
        grb_model.remove(remove_vars)
        grb_model.update()
        # grb_model.write('example/test_gurobi_removed.lp')


    def get_grb_solution(grb_model, reference, bound_type, eps=1e-5):
        refined = False
        if grb_model.status == 9: # Timed out. Get current bound.
            bound = bound_type(grb_model.objbound, reference)
            refined = abs(bound - reference) >= eps
        elif grb_model.status == 2: # Optimally solved.
            bound = grb_model.objbound
            refined = abs(bound - reference) >= eps
        elif grb_model.status == 15: # Found an lower bound >= 0 or upper bound <= 0, so this neuron becomes stable.
            bound = bound_type(1., -1.) * eps
            refined = True
        else:
            bound = reference
        return bound, refined, grb_model.status

    def solve_ub(model, v, out_ub, eps=1e-5):
        status_ub_r = -1  # Gurbo solver status.
        model.setObjective(v, grb.GRB.MAXIMIZE)
        model.reset()
        model.setParam('BestBdStop', -eps)  # Terminiate as long as we find a negative upper bound.
        # model.write(f'example/test_gurobi_ub.lp')

        try:
            model.optimize()
        except grb.GurobiError as e:
            handle_gurobi_error(e.message)
        vub, refined, status_ub = get_grb_solution(model, out_ub, min, eps=eps)
        return vub, refined, status_ub, status_ub_r

    def solve_lb(model, v, out_lb, eps=1e-5):
        status_lb_r = -1  # Gurbo solver status.
        model.setObjective(v, grb.GRB.MINIMIZE)
        model.reset()
        model.setParam('BestBdStop', eps)  # Terminiate as long as we find a positive lower bound.
        # model.write(f'example/test_gurobi_lb.lp')
        try:
            model.optimize()
        except grb.GurobiError as e:
            handle_gurobi_error(e.message)
        vlb, refined, status_lb = get_grb_solution(model, out_lb, max, eps=eps)
        return vlb, refined, status_lb, status_lb_r

    refine_time = time.time()
    model = MULTIPROCESS_MODEL.copy()
    l_id, n_id, var_name, solve_both, pre_relu_names, relu_names, final_name = candidate
    v = model.getVarByName(var_name)
    out_lb, out_ub = v.LB, v.UB
    neuron_refined = False
    eps = 1e-5
    v.LB, v.UB = -np.inf, np.inf
    model.update()

    if REMOVE_UNUSED:
        remove_unused_vars_and_constrs(model, var_name, {v_: k_ for k_, v_ in pre_relu_names.items()}, relu_names, final_name)

    if abs(out_lb) < abs(out_ub): # lb is tighter, solve lb first.
        vlb, refined, status_lb, status_lb_r = solve_lb(model, v, out_lb, eps=eps)
        neuron_refined = neuron_refined or refined
        if vlb <= 0 and solve_both: # Still unstable. Solve ub.
            vub, refined, status_ub, status_ub_r = solve_ub(model, v, out_ub, eps=eps)
            neuron_refined = neuron_refined or refined
        else: # lb > 0, neuron is stable, we skip solving ub.
            vub, status_ub, status_ub_r = out_ub, -1, -1
    else: # ub is tighter, solve ub first.
        vub, refined, status_ub, status_ub_r = solve_ub(model, v, out_ub, eps=eps)
        neuron_refined = neuron_refined or refined
        if vub >= 0 and solve_both: # Still unstable. Solve lb.
            vlb, refined, status_lb, status_lb_r = solve_lb(model, v, out_lb, eps=eps)
            neuron_refined = neuron_refined or refined
        else: # ub < 0, neuron is stable, we skip solving ub.
            vlb, status_lb, status_lb_r = out_lb, -1, -1

    if DEBUG:
        # print(model)
        if neuron_refined:
            print(f"Solving MIP for {v.VarName:<10}: [{out_lb:.6f}, {out_ub:.6f}]=>[{vlb:.6f}, {vub:.6f}] ({status_lb}, {status_ub}), time: {time.time()-refine_time:.4f}s, #vars: {model.NumVars}, #constrs: {model.NumConstrs}")
            sys.stdout.flush()

    return l_id, n_id, var_name, vlb, vub, neuron_refined, status_lb, status_ub



def print_tightened_bounds(name, olds, news):
    old_lowers, old_uppers = olds
    new_lowers, new_uppers = news

    old_lowers = old_lowers.flatten().detach().cpu()#.numpy()
    old_uppers = old_uppers.flatten().detach().cpu()#.numpy()

    new_lowers = new_lowers.flatten().detach().cpu()#.numpy()
    new_uppers = new_uppers.flatten().detach().cpu()#.numpy()

    print(f'[+] Layer: {name}')
    for i in range(len(old_lowers)):
        if old_lowers[i] * old_uppers[i] < 0:
            if (new_lowers[i] - old_lowers[i]).abs() > 1e-4 or (new_uppers[i] - old_uppers[i]).abs() > 1e-4:
                print(f'\t- neuron {i}: [{old_lowers[i]:.04f}, {old_uppers[i]:.04f}] => [{new_lowers[i]:.04f}, {new_uppers[i]:.04f}]')

class Tightener:

    def __init__(self, abstractor, objectives):
        self.abstractor = abstractor
        self.input_lowers = objectives.lower_bounds[0].clone().view(self.abstractor.input_shape).to(self.abstractor.device)
        self.input_uppers = objectives.upper_bounds[0].clone().view(self.abstractor.input_shape).to(self.abstractor.device)
        self.c_to_use = objectives.cs.clone().transpose(0, 1).to(self.abstractor.device)

        assert len(abstractor.net.relus) == len(abstractor.net.perturbed_optimizable_activations), print('[!] Error: Support ReLU only', len(abstractor.net.relus), len(abstractor.net.perturbed_optimizable_activations))
        self.pre_relu_names = {i: layer.inputs[0].name for (i, layer) in enumerate(abstractor.net.perturbed_optimizable_activations)}
        self.relu_names = {i: layer.name for (i, layer) in enumerate(abstractor.net.perturbed_optimizable_activations)}
        self.reset()


    def reset(self):
        self.black_list = []
        self.tightened_layers = []


    def select_layer(self):
        for l_id, l_name in self.pre_relu_names.items():
            if (l_id != 0) and (l_name not in self.tightened_layers):
                return l_name
        return random.choice(self.tightened_layers)


    # FIXME: only work with ReLU
    def __call__(self, domain_list, topk=64, largest=False, timeout=2.0, solve_both=False):
        # step 1: select domains
        worst_domains = domain_list.pick_out_worst_domains(len(domain_list), device='cpu')
        batch = len(list(worst_domains.lower_bounds.values())[0])
        # logger.debug(f'Tightening: {batch}')
        if batch == 0:
            return

        # worst bounds
        unified_lower_bounds = {k: v.min(dim=0).values.flatten() for k, v in worst_domains.lower_bounds.items()}
        unified_upper_bounds = {k: v.max(dim=0).values.flatten() for k, v in worst_domains.upper_bounds.items()}
        if os.environ.get('NEURALSAT_ASSERT'):
            assert all([(unified_lower_bounds[k] <= worst_domains.lower_bounds[k].flatten(1)).all() for k in worst_domains.lower_bounds])
            assert all([(unified_upper_bounds[k] >= worst_domains.upper_bounds[k].flatten(1)).all() for k in worst_domains.upper_bounds])
            assert len(self.pre_relu_names) == len(unified_lower_bounds)

        # print('lower_bounds:', [(k, v.shape, v.sum()) for k, v in unified_lower_bounds.items()])
        # print('upper_bounds:', [(k, v.shape, v.sum()) for k, v in unified_upper_bounds.items()])

        # step 2: select candidates
        unified_masks = {
            k: torch.where(unified_lower_bounds[k] * unified_upper_bounds[k] < 0)[0].numpy()
                for k in unified_lower_bounds if k != self.pre_relu_names[0] # skip first layer
        }

        unified_indices = [(k, vi) for k, v in unified_masks.items() for vi in v]

        unified_scores = torch.concat([
            torch.min(unified_upper_bounds[k][v].abs(), unified_lower_bounds[k][v].abs()).flatten()
                for k, v in unified_masks.items()
        ])

        assert unified_scores.numel() == len(unified_indices)

        if not len(unified_indices):
            return

        n_candidates = len(unified_indices)
        candidates = []
        selected_indices = unified_scores.topk(n_candidates, largest=largest).indices
        selected_layer = self.select_layer()
        # print('selected_layer:', selected_layer)

        for s_idx in selected_indices:
            l_id, n_id = unified_indices[s_idx]

            # 1st time, tighten only "selected_layer"
            if (l_id != selected_layer) and (selected_layer not in self.tightened_layers):
                continue

            var_name = f"lay{l_id}_{n_id}"
            if var_name in self.black_list:
                continue

            candidates.append((
                l_id,
                n_id,
                var_name,
                solve_both,
                self.pre_relu_names,
                self.relu_names,
                self.abstractor.net.final_name,
            ))

            # 1st time, tighten all neurons
            if selected_layer not in self.tightened_layers:
                continue

            # 2nd time, tighten topk neurons
            if len(candidates) == topk:
                break

        if selected_layer not in self.tightened_layers:
            self.tightened_layers.append(selected_layer)

        # step 3: rebuild mip model
        unified_bound_shapes = {k: v.size() for k, v in worst_domains.lower_bounds.items()}
        current_model = self.rebuild_mip_model(unified_lower_bounds, unified_upper_bounds, unified_bound_shapes)
        if current_model is None:
            return
        current_model.setParam('TimeLimit', timeout)

        global MULTIPROCESS_MODEL
        MULTIPROCESS_MODEL = current_model.copy()

        # step 4: tightening
        solver_result = []
        if len(candidates):
            with multiprocessing.Pool(min(len(candidates), os.cpu_count())) as pool:
                solver_result = pool.map(mip_solver_worker, candidates, chunksize=1)
        MULTIPROCESS_MODEL = None

        # step 5: update refined bounds
        unified_lower_bounds_refined = {k: v.clone() for k, v in unified_lower_bounds.items()}
        unified_upper_bounds_refined = {k: v.clone() for k, v in unified_upper_bounds.items()}
        unstable_to_stable_neurons = []
        num_neuron_refined = 0
        for l_id, n_id, var_name, vlb, vub, neuron_refined, s_lb, s_ub in solver_result:
            if neuron_refined:
                num_neuron_refined += 1
                unified_lower_bounds_refined[l_id][n_id] = max(unified_lower_bounds[l_id][n_id], vlb)
                unified_upper_bounds_refined[l_id][n_id] = min(unified_upper_bounds[l_id][n_id], vub)
                if vlb > 0:
                    unstable_to_stable_neurons.append((l_id, n_id, 1.0))
                elif vub < 0:
                    unstable_to_stable_neurons.append((l_id, n_id, -1.0))
                # print(f'neuron[{l_id}][{n_id}]: [{unified_lower_bounds[l_id][n_id]:.06f}, {unified_upper_bounds[l_id][n_id]:.06f}] => [{unified_lower_bounds_refined[l_id][n_id]:.06f}, {unified_upper_bounds_refined[l_id][n_id]:.06f}]')

            # add neurons to blacklist
            # self.black_list.append(var_name)

        logger.debug(f'[Stabilize] layer="{selected_layer}", #candidates={len(candidates)}, #total={len(unified_indices)}, #tightened={num_neuron_refined}, #stabilized={len(unstable_to_stable_neurons)}, #blacklisted={len(self.black_list)}')

        # print('lower_bounds:', [(k, v.shape, v.sum()) for k, v in unified_lower_bounds_refined.items()])
        # print('upper_bounds:', [(k, v.shape, v.sum()) for k, v in unified_upper_bounds_refined.items()])

        # step 6: update domains bounds
        class TMP:
            pass

        refined_domain = TMP()
        refined_domain.lower_bounds = unified_lower_bounds_refined
        refined_domain.upper_bounds = unified_upper_bounds_refined

        domain_list.update_refined_bounds(refined_domain)

        return


    def rebuild_mip_model(self, refined_lower_bounds, refined_upper_bounds, bound_shapes):
        intermediate_layer_bounds = {}
        assert len(bound_shapes) == len(refined_lower_bounds) == len(refined_upper_bounds)

        for l_name, shape in bound_shapes.items():
            intermediate_layer_bounds[l_name] = [
                refined_lower_bounds[l_name].to(self.abstractor.device).view(1, *shape[1:]),
                refined_upper_bounds[l_name].to(self.abstractor.device).view(1, *shape[1:])
            ]
        try:
            self.abstractor.build_lp_solver(
                model_type='mip',
                input_lower=self.input_lowers,
                input_upper=self.input_uppers,
                c=self.c_to_use,
                refine=False,
                intermediate_layer_bounds=intermediate_layer_bounds,
            )
        except AttributeError:
            return None
        except:
            raise NotImplementedError()

        current_model = self.abstractor.net.model.copy()
        current_model.setParam('Threads', 1)
        current_model.setParam('MIPGap', 0.01)
        current_model.setParam('MIPGapAbs', 0.01)
        current_model.update()

        return current_model



import random
import torch
import copy
import time

import math

from tqdm import trange


def rand(total_samples, num_samples, device='cpu'):
    if num_samples >= total_samples:
        return torch.Tensor(range(total_samples), device=device)
    return torch.Tensor(random.sample(range(total_samples), num_samples), device=device)



# print(rand(10, 5))
# print(rand(10, 15))

# X = torch.zeros(10)
# print(torch.empty_like(X).uniform_())


def test_solver():
    from solver.sat_solver import SATSolver
    import time
    
    init_alls = []
    
    if 0:
        for i in range(5000):
            init_alls.append(SATSolver([[_ for _ in range(i)]]))
            # init_alls.append(SATSolver([frozenset({i})]))
            
        tic = time.time()
        alls = init_alls + copy.deepcopy(init_alls)
        print('done', time.time() - tic)
        return
       
    init_alls = []
    for i in range(3):
        init_alls.append(SATSolver([[i+1], [i+1, i + 2], [i+1, -(i+3)]]))
        
    tic = time.time()
    alls = init_alls + copy.deepcopy(init_alls)
    # alls = init_alls + init_alls
    print('done', time.time() - tic)
    
    
    print('-----------all------------')
    [_.print_stats() for _ in alls]
    
    batch = 2
    picks = alls[-batch:]
    alls = alls[:-batch]
    
    
    print('-----------pick------------')
    [_.print_stats() for _ in picks]
    
    
    print('-----------remain------------')
    [_.print_stats() for _ in alls]
    
    
    # for i in range(5, 8):
    #     alls.append(SATSolver([frozenset({i})]))
    
    print('-----------bcp------------')
    [_.bcp() for _ in picks]
        
    print('-----------add------------')
    [alls.append(_) for _ in picks]
    
    
    print('-----------all------------')
    [_.print_stats() for _ in alls]
    
    
    
    
def test_small():
    size = 3
    a = {}
    for i in range(size):
        a[i] = [list(range(_)) for _ in range(i+2)]
        a[i] = list(range(i+2))
    # b = {}
    b = {k: v.copy() for k, v in a.items()}
    b[1].remove(1)
    print(a)
    print(b)
    
    tic = time.time()
    b = copy.deepcopy(a)
    print('copy dict in:', time.time() - tic)
    
    # a = [[1, 2, 3], [3, 4, 5]]
    # b = a[:]
    # b[0].remove(1)
    
    a = []
    for i in range(size):
        a.append(list(range(i+2)))
    
    tic = time.time()
    b = [_[:] for _ in a]
    b[1].remove(1)
    # print('copy list in:', time.time() - tic)
    print(a)
    print(b)
    
    
def test():
    
    a = torch.tensor([1, 2, 3, 4, 5])  
    b = torch.tensor([2, 3])  
    print(a)
    print(torch.where(a == b))
    a = a[a != b]
    print(a)



# @torch.jit.script
def haioc_old(data: torch.Tensor, xs: torch.Tensor, inplace: bool = False) -> torch.Tensor:
    for i in range(xs.size(0)):
        remove_idx = torch.where(data == xs[i])[0]
        remain_mask = torch.ones(len(data), dtype=torch.bool)
        remain_mask[remove_idx] = False
        data = data[remain_mask]
        if not inplace:
            data = torch.where(data == -xs[i], 0, data)
        else:
            data[data == -xs[i]] = 0
    return data


# @torch.jit.script
def haioc(data: torch.Tensor, xs: torch.Tensor, inplace: bool = False) -> torch.Tensor:
    remain_mask = torch.ones(data.size(0), dtype=torch.bool, device=data.device)
    for i in range(xs.size(0)):
        remain_mask &= data.ne(xs[i]).all(1)
    data = data[remain_mask]
    zero_mask = torch.zeros(data.size(), dtype=torch.bool, device=data.device)
    for i in range(xs.size(0)):
        zero_mask |= data.eq(-xs[i])
    if not inplace:
        data = torch.where(zero_mask, 0, data)
    else:
        data[zero_mask] = 0
    return data


class TimeMeter:
    def __init__(self):
        self.avg = self.count = 0
        self.start = self.end = None

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = time.time()
        elapsed_time = self.end - self.start
        self.avg = (self.avg * self.count + elapsed_time) / (self.count + 1)
        self.count += 1

    def reset(self):
        self.avg = self.count = 0
        self.start = self.end = None

    @property
    def fps(self):
        return 1. / self.avg if self.avg else math.nan


def main(n_trials=100, inplace=False):
    size = (5000, 50)
    n = 3000
    data = torch.randperm(size[0] * size[1]).sub_(size[0] * size[1] // 2).view(size[0], size[1]).int()
    xs = torch.arange(0, n).int()

    time_meter = TimeMeter()

    # warmup if using jit scripting:
    if isinstance(haioc, torch.jit.ScriptFunction):
        haioc(data, xs, inplace=inplace)

    time_meter.reset()
    pbar = trange(n_trials)
    for i in pbar:
        torch.manual_seed(i)
        with time_meter:
            haioc_old(data, xs, inplace=inplace)
        pbar.set_description(f'[old] average_fps={time_meter.fps:.05f}')

    time_meter.reset()
    pbar = trange(n_trials)
    for i in pbar:
        torch.manual_seed(i)
        with time_meter:
            haioc(data, xs, inplace=inplace)
        pbar.set_description(f'[new] average_fps={time_meter.fps:.05f}')

def extract_log():
    import os
    import re
    
    root = './mnistfc_hard_results/'
    with open('valid_instances.csv', 'w') as fp:
        for file in os.listdir(root):
            file_path = os.path.join(root, file)
            content = open(file_path).read().strip()
            if ('unsat' not in content) and ('timeout' not in content):
                continue
            
            net, spec = file.split('_[spec]_')
            net = net.split('[net]_')[1]
            spec = spec.split('.log')[0]
            # print(net, spec)
            # if 'unsat' in content:
            if 'timeout,' in content:
                last_iter_log = content.split('\n')[-3]
                last_bound = float(re.findall(r'Bound: -([0-9\.]+)', last_iter_log)[0])
                # print(last_bound)
                if last_bound >= 5.0:
                    continue
                print(last_iter_log)
            print(file_path)
                
            print(f'onnx/{net}.onnx,vnnlib/{spec}.vnnlib,3600', file=fp)

if __name__ == '__main__':
    # main(100, True)
    extract_log()




from functools import reduce
import operator

import torch
from torch import nn

from ..operations.base import Operator
from ..utils import assign_values_to_dim, get_selection, PRINT_DEBUG


def prod(x):
    return reduce(operator.mul, x, 1)

# https://github.com/KaidiXu/onnx2pytorch/

class Reshape(Operator):
    """
    In the initial pass it stores the initial_input_shape.
    It uses it to infer the new reshape value from a
    smaller pruned input in the following passes.
    """

    def __init__(self, enable_pruning, shape=None, keep_size=True, quirks=None):
        super().__init__()
        self.enable_pruning = enable_pruning
        self.shape = shape
        self.initial_input_shape = None
        self.feature_dim = -1
        self.input_indices = None
        self.placeholder = None
        self.keep_size = keep_size
        self.quirks = {} if quirks is None else quirks
        assert isinstance(self.quirks, dict)

    def forward(self, input: torch.Tensor, shape=None):
        shape = shape if shape is not None else self.shape
        if shape[0] == 1 and len(shape) in [2, 3, 4, 5] and self.quirks.get("fix_batch_size") is True:
            incomplete_indices = (shape == -1).nonzero()
            assert incomplete_indices.numel() <= 1, "at most one dimension can be -1 in reshape"
            if incomplete_indices.numel() == 1 and shape[0] != -1:
                # Have a -1 shape not at the batch dimension.
                incomplete_loc = incomplete_indices[0, 0].item()
                # Need to compute the actual shape if we already have a -1.
                incomplete_shape = -1 * torch.prod(shape[1:])
                inferred_shape = prod(input.shape[1:]) // incomplete_shape
                shape[incomplete_loc] = inferred_shape
            shape[0] = -1
            if self.initial_input_shape is None:
                print("Enabling quirks for Reshape operation: fix the first " "dimension shape to be -1 to support batchsize != 1.")
                print(f"input shape {input.shape}, new shape is {shape}.")
        else:
            # This raises RuntimeWarning: iterating over a tensor.
            # FIXME: this looks not right.
            # if the first dim is batch size, merge the batch size with the channel dim
            if (torch.prod(torch.tensor(input.shape)) != torch.prod(shape) and len(input.size()) == len(shape) + 1
                    and torch.prod(torch.tensor(input.shape)) == input.shape[0] * torch.prod(shape)
                    and self.quirks.get("merge_batch_size_with_channel") is True):
                shape = torch.tensor([input.shape[0] * input.shape[1] // 2] + shape.tolist()[1:], device=shape.device)
            shape = [x if x != 0 else input.size(i) for i, x in enumerate(shape)]
        if not self.enable_pruning:
            return torch.reshape(input, tuple(shape))

        inp_shape = torch.tensor(input.shape)
        if self.initial_input_shape is None:
            self.initial_input_shape = inp_shape
        elif len(shape) == 2 and shape[-1] == -1:
            pass
        elif torch.equal(self.initial_input_shape, inp_shape):
            # input's shape did not change
            pass
        elif self.input_indices is not None:
            self.placeholder *= 0
            selection = get_selection(self.input_indices, self.feature_dim)
            self.placeholder[selection] += input
            input = self.placeholder
        elif torch.prod(inp_shape) == torch.prod(torch.tensor(shape)):
            # If input's shape changed but shape changed to account for this,
            # no additional work is needed.
            # This happens when shape is dynamically computed by the network.
            pass
        else:
            # If input's shape changed but shape has not accounted for this,
            # the reshaped shape must change as well.
            c = torch.true_divide(inp_shape, self.initial_input_shape)
            if len(c) < len(shape) and shape[0] == 1:
                c = torch.cat((torch.tensor([1]), c))
            shape = (c * torch.tensor(shape)).to(int)
        return torch.reshape(input, tuple(shape))

    def set_input_indices(self, input):
        input_shape = input[0].shape
        if self.feature_dim < 0:
            self.feature_dim += len(input_shape)
        axis = self.get_axis(input_shape, self.feature_dim)
        mask = input[0] != 0
        s = mask.sum(axis=tuple(axis))
        mask = s != 0
        (non_zeros,) = torch.where(mask)
        self.input_indices = non_zeros
        self.placeholder = nn.Parameter(
            torch.zeros(*self.initial_input_shape, device=input[0].device, dtype=input[0].dtype),
            requires_grad=False,
        )

    def extra_repr(self) -> str:
        return "shape={}".format(self.shape)



import warnings

import torch
from torch import nn


class Argmax(nn.Module):
    def __init__(self, axis=None):
        self.axis = axis
        super().__init__()

    def forward(self, data):
        return torch.argmax(data, dim=self.axis)
